import numpy as np
from numpy import ma
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.animation as anm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import colors
from matplotlib.colors import LinearSegmentedColormap

from skimage.util import montage
from skimage.filters import rank
from skimage import morphology
from skimage import exposure
from skimage import measure
from skimage import filters
from skimage import io
from skimage import transform
from skimage import registration

from scipy import ndimage
from scipy import signal
from scipy import stats
from scipy import ndimage as ndi

from ..utils import masking
from ..utils import plot


class wf_x2_m2():
    def __init__(self, img_path:str, img_name:str,
                ch_order:dict[str:int], use_gauss:bool=False, wf_sigma:float=.5, border_crop:int=0,
                **kwargs):
        """ Class is designed to store experiment data from wide-field imaging
        using two different excitation wavelengths and two emission channels (__2 eXcitation & 2 eMission__).
        It's specifically optimized for the results of individual neuron imaging.

        Parameters
        ----------
        img_path: str
            path to the image series TIFF file
        img_name: str
            recording name
        ch_order: dict
            indexes of 1st and 2nd fluorescence proteins channels,
            if fluorescence proteins doesn't overlap (`{'fp1': index, 'fp2': index}`)
        use_gauss: boolean, optional
            if `True` Gaussian filter will be applied to input image series
        wf_sigma: float, optional
            sigma value for Gaussian filter applied to input image series
        border_crop: int (0)
            image crop size in pixels,
            this amount of pixels will be deleted from sides on each frame
            
        Attributes
        ----------
        img_raw: ndarray [t,x,y,c]
            raw input image series
        img_name: str
            registration name
        fp1_img: ndarray [t,x,y]
            1st fluorescence protein image series
            with photobleaching correction and Gaussian filtering (if `use_gauss` is `True`)
        fp1_img_raw: ndarray [t,x,y]
            1st fluorescence protein image series,
            without photobleaching correction and Gaussian filtering
        fp1_img_corr: ndarray [t,x,y]
            1st fluorescence image series
            with photobleaching correction
        fp1_mean_img_raw: ndarray [x,y]
            1st fluorescence protein pixel-wise mean image
        fp2_img: ndarray [t,x,y]
            2nd fluorescence protein image series,
            with photobleaching correction and Gaussian filtering (if `use_gauss` is `True`) 
        fp2_img_raw: ndarray [t,x,y]
            2nd fluorescence protein image series,
            without photobleaching correction and Gaussian filtering
        fp2_img_corr: ndarray [t,x,y]
            2nd fluorescence protein image series
            with photobleaching correction
        fp2_mean_img_raw: ndarray [x,y]
            2nd fluorescence protein pixel-wise mean image
        proc_mask: ndarray [x,y]
            cell processes boolean mask, extended (created with `utils.masking.proc_mask()`)
        narrow_proc_mask: ndarray [x,y]
            cell processes boolean mask, unextended (created with `utils.masking.proc_mask()`)


        """
        self.img_raw = io.imread(img_path)
        if border_crop !=0:  # optional crop of image series
            y,x = self.img_raw.shape[1:3]
            self.img_raw = self.img_raw[:,border_crop:y-border_crop,border_crop:x-border_crop,:]

        self.img_name = img_name
        self.ch_order = ch_order
        self.gauss_sigma = gauss_sigma

        # primary image extraction
        self.fp1_img_raw = self.img_raw[:,:,:,ch_order['fp1']]
        self.fp2_img_raw = self.img_raw[:,:,:,ch_order['fp2']]

        self.fp1_mean_img_raw = np.mean(self.fp1_img_raw, axis=0)
        self.fp2_mean_img_raw = np.mean(self.fp2_img_raw, axis=0)

        # mask creation
        self.proc_mask = masking.proc_mask(self.fp2_mean_img_raw,
                                           ext_fin_mask=True, **kwargs)
        self.narrow_proc_mask = masking.proc_mask(self.fp2_mean_img_raw,
                                                  ext_fin_mask=False, **kwargs)

        # photobleaching correction and bluring
        self.fp1_img_corr,self.fb1_pbc,self.fb1_pb_r = masking.pb_exp_correction(input_img=self.fp1_img_raw,
                                                                                 mask=self.proc_mask)
        
        self.fp1_back_mean = np.mean(self.fp1_img_corr, axis=(1,2), where=~self.proc_mask)
        self.fp1_back_mean = self.fp1_back_mean.reshape(-1, 1, 1)
        self.fp1_img_corr = self.fp1_img_corr - self.fp1_back_mean
        self.fp2_img_corr,self.fb2_pbc,self.fb2_pb_r = masking.pb_exp_correction(input_img=self.fp2_img_raw,
                                                                                 mask=self.proc_mask)
        self.fp2_back_mean = np.mean(self.fp2_img_corr, axis=(1,2), where=~self.proc_mask)
        self.fp2_back_mean = self.fp2_back_mean.reshape(-1, 1, 1)
        self.fp2_img_corr = self.fp2_img_corr - self.fp2_back_mean

        if use_gauss:
            self.fp1_img = np.asarray([filters.gaussian(frame, sigma=self.gauss_sigma) \
                                       for frame in self.fp1_img_corr])
            self.fp2_img = np.asarray([filters.gaussian(frame, sigma=self.gauss_sigma) \
                                       for frame in self.fp2_img_corr])
        else:
            self.fp1_img = self.fp1_img_corr
            self.fp2_img = self.fp2_img_corr


        # # img maskings
            # mask_arr: ndarray [t,x,y]
            #     cell processes mask array, series of `proc_mask`
            # masked_fp1_img: ndarray [t,x,y]
            #     1st fluorescence protein image series masked frame-by-frame with `proc_mask`,
            #     out of mask pixels set as 0    
            # masked_fp2_img: ndarray [t,x,y]
            #     2nd fluorescence protein image series masked frame-by-frame with `proc_mask`,
            #     out of mask pixels set as 0
        # self.mask_arr = np.asarray([z+self.proc_mask for z in np.zeros_like(self.fp1_img)], dtype='bool')
        # self.masked_fp1_img = np.copy(self.fp1_img)
        # self.masked_fp1_img[~self.mask_arr] = 0

        # self.masked_fp2_img = np.copy(self.fp2_img)
        # self.masked_fp2_img[~self.mask_arr] = 0


        # PB CORR VARIANTS
        # pb corr with constant val, old version with zeros background
        # self.corr_fp1_img = np.asarray([img * (np.sum(np.mean(self.masked_fp1_img[:2], axis=0)/np.sum(img))) \
        #                                   for img in self.masked_fp1_img])
        # self.corr_fp2_img = np.asarray([img * (np.sum(np.mean(self.masked_fp2_img[:2], axis=0)/np.sum(img))) \
        #                                   for img in self.masked_fp2_img])
        
        # pb corr with constant val, without zeros background
        # self.corr_fp1_img = np.asarray([self.fp1_img[img_i] * (np.sum(np.mean(self.masked_fp1_img[:2], axis=0)/np.sum(self.masked_fp1_img[img_i]))) \
        #                                   for img_i in range(self.masked_fp1_img.shape[0])])
        # self.corr_fp2_img = np.asarray([self.fp2_img[img_i] * (np.sum(np.mean(self.masked_fp2_img[:2], axis=0)/np.sum(self.masked_fp2_img[img_i]))) \
        #                                   for img_i in range(self.masked_fp2_img.shape[0])])


    def get_all_channels(self):
        """ Returns all individual channels blurred with Gaussian filter (with sigma `wf_sigma`).
        
        Could be useful for FRET calculation.

        Returns
        -------
        ch0_img: ndarray [t,x,y]
            image series for channel 0
        ch1_img: ndarray [t,x,y]
            image series for channel 0
        ch2_img: ndarray [t,x,y]
            image series for channel 0
        ch3_img: ndarray [t,x,y]
            image series for channel 0

        """
        # chennels split
        ch0_img = np.asarray([filters.gaussian(frame, sigma=self.gauss_sigma) \
                              for frame in self.img_raw[:,:,:,0]])  # 435-fp1  DD
        ch1_img = np.asarray([filters.gaussian(frame, sigma=self.gauss_sigma) \
                              for frame in self.img_raw[:,:,:,1]])  # 435-fp2  DA
        ch2_img = np.asarray([filters.gaussian(frame, sigma=self.gauss_sigma) \
                              for frame in self.img_raw[:,:,:,2]])  # 505-fp1  AD
        ch3_img = np.asarray([filters.gaussian(frame, sigma=self.gauss_sigma) \
                              for frame in self.img_raw[:,:,:,3]])  # 505-fp2  AA

        return ch0_img, ch1_img, ch2_img, ch3_img

    def ch_pic(self):
        """ Plotting of registration control image

        """
        v_min = np.min(self.img_raw)
        v_max = np.max(self.img_raw) * .5

        fp1_der_w, fp1_der_p = masking.series_derivate(input_img=self.fp1_img_corr,
                                                       mask=self.proc_mask,
                                                       left_win=2, space=4, right_win=2)
        fp2_der_w, fp2_der_p = masking.series_derivate(input_img=self.fp2_img_corr,
                                                       mask=self.proc_mask,
                                                       left_win=2, space=4, right_win=2)

        plt.figure(figsize=(10,10))
        ax0 = plt.subplot(231)
        ax0.set_title('fp1')
        ax0.imshow(self.fp1_mean_img_raw, cmap='jet')
        ax0.axis('off')

        ax1 = plt.subplot(232)
        ax1.set_title('fp2')
        img1 = ax1.imshow(self.fp2_mean_img_raw, cmap='jet')
        img1.set_clim(vmin=v_min, vmax=v_max)
        div1 = make_axes_locatable(ax1)
        cax1 = div1.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img1, cax=cax1)
        ax1.axis('off')

        ax4 = plt.subplot(233)
        ax4.imshow(self.fp2_mean_img_raw, cmap='jet')
        ax4.imshow(ma.masked_where(~self.narrow_proc_mask, self.narrow_proc_mask),
                   cmap='Greys', alpha=.4)
        ax4.imshow(ma.masked_where(~self.proc_mask, self.proc_mask),
                   cmap='Greys', alpha=.25)
        ax4.set_title('processes mask')
        ax4.axis('off')

        ax2 = plt.subplot(413)
        ax2.plot(np.mean(self.fp1_img_raw, axis=(1,2), where=self.proc_mask),
                 label='fp1 raw', color='k')
        ax2.plot(np.mean(self.fp1_img_corr, axis=(1,2), where=self.proc_mask),
                 label='fp1 corrected', linestyle='--', color='k')
        ax2.plot(np.mean(self.fp2_img_raw, axis=(1,2), where=self.proc_mask),
                 label='fp2 raw', color='r')
        ax2.plot(np.mean(self.fp2_img_corr, axis=(1,2), where=self.proc_mask),
                 label='fp2 corrected', linestyle='--', color='r')
        ax2.set_xlabel('Frame num')
        ax2.set_ylabel('Int, a.u.')
        ax2.set_title('Int. profile')
        ax2.legend()

        ax3 = plt.subplot(414)
        ax3.plot(fp2_der_p, linestyle='--', color='r')
        ax3.plot(fp1_der_p, linestyle='--', color='k')
        ax3.plot(fp2_der_w, color='r')
        ax3.plot(fp1_der_w, color='k')
        ax3.set_xlabel('Frame num')
        ax3.set_ylabel('Norm. der.')
        ax3.set_title('Der. profile')
        ax3.legend()

        plt.suptitle(self.img_name)
        plt.tight_layout()
        plt.show()

    
    def processes_mask_pic(self):
        """ Plotting of cell processes mask (narrow and extended)
        overlay with control fluorescence image (`fp2_mean_img_raw`).

        """
        plt.figure(figsize=(10,190))
        ax0 = plt.subplot()
        ax0.imshow(self.fp2_mean_img_raw, cmap='jet')
        ax0.imshow(ma.masked_where(~self.narrow_proc_mask, self.narrow_proc_mask),
                   cmap='Greys', alpha=.4)
        ax0.imshow(ma.masked_where(~self.proc_mask, self.proc_mask),
                   cmap='Greys', alpha=.25)
        ax0.axis('off')

        plt.tight_layout()
        plt.show()


    def hist_pic(self):
        """ Plotting of histogram for individual channels
        (for channel mean intensity frame).
        
        """

        ch0_ctrl = np.mean(self.img_raw[:,:,:,0], axis=0)
        ch1_ctrl = np.mean(self.img_raw[:,:,:,1], axis=0)
        ch2_ctrl = np.mean(self.img_raw[:,:,:,2], axis=0)
        ch3_ctrl = np.mean(self.img_raw[:,:,:,3], axis=0)

        plt.figure(figsize=(20,10))
        ax0 = plt.subplot()
        ax0.hist(ch0_ctrl.ravel(), bins=256, alpha=.5, label='Ch 0 (fp1-435)', color='r')
        ax0.hist(ch1_ctrl.ravel(), bins=256, alpha=.5, label='Ch 1 (fp2-435)', color='g')
        ax0.hist(ch2_ctrl.ravel(), bins=256, alpha=.5, label='Ch 2 (fp1-505)', color='y')
        ax0.hist(ch3_ctrl.ravel(), bins=256, alpha=.5, label='Ch 3 (fp2-505)', color='b')
        ax0.legend()

        plt.show()


import time
import autograd.numpy as np
from numpy import linalg as la
from scipy.stats import norm, lognorm, uniform
from scipy.integrate import quad, quadrature
from pathos.multiprocessing import ProcessingPool
from pathos.helpers import cpu_count
		
		
def diag_gaussian_2d_pdf(x, mean, cov):
	"""
	Returns the pdf of a 2d Gaussian where each dimension is separable.
	i.e. the covariance matrix must be diagonal. Should be faster
	than using matrix multiplication for a general 2D covariance.
	"""
	
	assert np.prod(cov[0, 1]*cov[1, 0]) == 0, 'Need diagonal covariance'
	
	exp_arg = -0.5*(1/cov[0, 0]*(x[0] - mean[0])**2 \
		  + 1/cov[1, 1]*(x[1] - mean[1])**2)
	
	return np.exp(exp_arg)/2*np.pi/np.sqrt(cov[0, 0]*cov[1, 1])
	
class norm_pdf():		
	"""
	Class for pdf of a normal distribution given fixed SD and mean.
	"""
	
	def __init__(self, sigma, fixed_mode=None):
		self.sigma = sigma
		self.fixed_mode = fixed_mode
		
	def _f(self, x):
		return 1/np.sqrt(2*np.pi*self.sigma**2)*\
			   np.exp(-(x - self.mean)**2/2/self.sigma**2)
		
	def pdf(self, p_init):
	
		if self.fixed_mode is None:
			self.mean = p_init
		else:
			self.mean = self.fixed_mode
		
		return self._f
		
class lognorm_pdf():
	"""
	Class for pdf of a lognormal distribution given fixed SD and mode.
	"""
	
	def __init__(self, sigma, fixed_mode=None):
		self.sigma = sigma
		self.fixed_mode = fixed_mode
		
	def _f(self, x):
		return 1/(x*self.sigma*np.sqrt(2*np.pi))*\
			   np.exp(-(np.log(x) - self.mean)**2/(2*self.sigma**2))
		
	def pdf(self, p_init):
	
		# Assume p_init is mode of lognormal -- convert to mean
		if self.fixed_mode is None:
			self.mean = np.log(p_init) + self.sigma**2.0
		else:
			self.mean = np.log(self.fixed_mode) + self.sigma**2.0
		
		return self._f
		
class standard_uniform_pdf():
	"""
	Class for pdf of standard uniform distribution
	"""
	
	def __init__(self):
		pass
		
	def _f(self, x):
		return 1.*(x > 0.)*(x <= 1.)
	
	def pdf(self, p_init):
		
		return self._f
		
def kalman_filter(Q, H, R, y, m_0, P_0, t_vec):
	"""
	Calculates posterior distribution at points in t_vec, for Gaussian 
	measurements and a linear-Gaussian model dynamicsl, using the Kalman
	filter update equations.
	"""
	
	N = len(t_vec)
	median_delta_t = np.median(np.diff(t_vec))
	d_t_vec = 0
	
	# Indexed assignment is not allowed by autograd so must append to lists
	m_pred = []
	P_pred = []
	m = []
	P = []
	S = []
	v = []
	
	m_k = m_0
	P_k = P_0
		
	for k in range(N):
		
		m_pred_k = m_k

		# Q is constant
		if len(Q.shape) == 2:
			P_pred_k = P_k + Q
		
		# If Q is time-dependent
		else:
		
			# If sampling interval large, don't use recent information
			if (k == 0) or (d_t_vec > 2*median_delta_t):
				P_pred_k = P_k + Q[0]
			else:
				P_pred_k = P_k + Q[k - 1]
		
		v_k = y[k] - np.dot(H[k], m_pred_k)
		S_k = np.dot(np.dot(H[k], P_pred_k), H[k].T) + R[k]
		K_k = np.dot(np.dot(P_pred_k, H[k].T), np.linalg.inv(S_k))
		m_k = m_pred_k + np.dot(K_k, v_k)
		P_k = P_pred_k - np.dot(np.dot(K_k, H[k]), P_pred_k)
		
		m_pred.append(m_pred_k)
		P_pred.append(P_pred_k)
		m.append(m_k)
		P.append(P_k)
		S.append(S_k)
		v.append(v_k)
		t_vec_increase = t_vec[min(k + 1, N - 1)] - t_vec[k]
	
	return np.array(m_pred), np.array(P_pred), np.array(m), \
			 np.array(P), np.array(S), np.array(v)

def RTS_smoother(m_pred, P_pred, m, P):
	"""
	Calculates smoothed posterior at points in t_vec, for Gaussian 
	measurements and a linear-Gaussian model dynamicsl, using the 
	predicted distributions of a Kalman filter through the dataset.
	"""
	
	N = m.shape[0]
	
	ms = [m[-1]]
	Ps = [P[-1]]
	
	ms_k = ms[0]
	Ps_k = Ps[0]
	
	# Iterate from penultimate index to 0
	for k in np.arange(N - 2, -1, -1):
		G_k = np.dot(P[k], np.linalg.inv(P_pred[k + 1]))
		ms_k = m[k] + np.dot(G_k, ms_k - m_pred[k + 1])
		Ps_k = P[k] + np.dot(G_k, np.dot(Ps_k - P_pred[k + 1], G_k.T))
		ms.append(ms_k)
		Ps.append(Ps_k)
	
	# Arrays are defined backward
	ms = ms[::-1]
	Ps = Ps[::-1]
	
	return np.array(ms), np.array(Ps)

def zeus_mcmc(log_prob, init, num_wkrs, num_samples, burnin=0, all_cores=True,
			  d_init=0.1):
	"""
	Slice sampling to generate MCMC samples of the log of an unnormalized
	probability distribution. Can run on parallel cores of a node.
	"""
	
	print ('MCMC sampling of posterior distribution ...')
	timenow = time.time()
	num_dim = len(init)
	init_tiled = np.zeros((num_wkrs, num_dim))
	
	import zeus	
	def run_parallel(seed):
		
		np.random.seed(seed)
		for i in range(num_wkrs):
			inits = np.random.normal(init, d_init, num_dim)
			while not np.isfinite(log_prob(inits)):
				inits = np.random.normal(init, d_init, num_dim)
			init_tiled[i] = inits
		samplers = zeus.EnsembleSampler(num_wkrs, num_dim, log_prob,
					 light_mode=True, check_walkers=False) 
		samplers.run_mcmc(init_tiled, num_samples)
		samples = samplers.get_chain(flat=True, discard=burnin)
		
		return samples
	
	if all_cores:
		pool = ProcessingPool()
		chains = pool.map(run_parallel, range(cpu_count()))
		all_samples = np.concatenate(chains, axis=0)
	else:
		all_samples = run_parallel(0)
	
	print ('Time elapsed:', time.time() - timenow)
	return all_samples
	
def non_Gauss_1D_filtering(xs, Q_pdf, LHs, m_0, P_0, t_vec):
	"""
	Generates the posterior distribution given i) a Gaussian prior, ii) the
	observation likelihoods at all points, as a funcition of discrete points 
	x, and iii) a function that returns pdf of the process noise from x_{t} 
	to x_{t + 1}. The pdf is a function of (x_{t+1} - x_t}, which can be 
	non-Gaussian. This is for a 1D state space.
	"""
	
	dx = xs[1] - xs[0]
	N = len(t_vec)
	int_N = len(xs)
	assert LHs.shape[0] == N, 'Likelihood arr must be (# timesteps, # xs)'
	assert LHs.shape[1] == int_N, 'Likelihood arr must be (# timesteps, # xs)'
	
	# Probability of transition from x_n (row) to x_{n + 1} (column). 
	Q_mat = Q_pdf(xs[:, None] - xs[None, :])
	
	prior = norm_pdf(P_0**0.5).pdf(m_0)(xs)
	pred_probs = []
	post_probs = []
	
	for i in range(N):
		
		# Predictive step; yes this is the right order; since rows are x_n
		p_tilde = np.dot(prior, Q_mat)
		p_tilde /= np.sum(dx*p_tilde)
		
		# Update step 
		f_tilde = LHs[i]*p_tilde
		f_tilde /= np.sum(dx*f_tilde)
		
		# Prior for next timestep is current posterior
		prior = f_tilde
		pred_probs.append(p_tilde)
		post_probs.append(f_tilde)
	
	dist = dict()
	dist['pred_probs'] = np.array(pred_probs)
	dist['post_probs'] = np.array(post_probs)
	dist['Q_mat'] = Q_mat
	dist['xs'] = xs
	dist['dx'] = dx
	
	return dist
	
def non_Gauss_1D_smoothing(dist):
	
	Q_mat = dist['Q_mat']
	p_tildes = dist['pred_probs']
	f_tildes = dist['post_probs']
	N = len(p_tildes)
	dx = dist['dx']
	
	# To hold backwards smoothed posteriors at each timepoint
	smth_post_probs = []
	cdfs = []
	s_tilde = f_tildes[-1]
	smth_post_probs.append(s_tilde)
	cdfs.append(np.cumsum(dx*s_tilde))
	
	for i in range(N - 2, -1, -1):
		
		f_tilde = f_tildes[i]
		p_tilde = p_tildes[i + 1]
		p_tilde[p_tilde < 1e-10] = 1e-10
		s_tilde = f_tilde*np.dot(s_tilde/p_tilde, Q_mat)
		s_tilde /= np.sum(s_tilde*dx)
		smth_post_probs.append(s_tilde)
		cdfs.append(np.cumsum(dx*s_tilde))
	
	dist['smth_post_probs'] = smth_post_probs[::-1]
	dist['cdfs'] = cdfs[::-1]
	
	return dist
	

def nearest_PD(A):
	"""
	Find the nearest positive-definite matrix to input
	"""

	B = (A + A.T)/2
	_, s, V = la.svd(B)
	H = np.dot(V.T, np.dot(np.diag(s), V))
	A2 = (B + H)/2
	A3 = (A2 + A2.T)/2

	if is_PD(A3):
		return A3

	spacing = np.spacing(la.norm(A))
	I = np.eye(A.shape[0])
	k = 1
	
	while not is_PD(A3):
		mineig = np.min(np.real(la.eigvals(A3)))
		A3 += I * (-mineig * k**2 + spacing)
		k += 1

	return A3


def is_PD(B):
	"""
	Returns true when input is positive-definite, via Cholesky
	"""
	
	try:
		la.cholesky(B)
		return True
		
	except la.LinAlgError:
		return False


import numpy as np
from numpy import ma
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.animation as anm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import colors
from matplotlib.colors import LinearSegmentedColormap

import plotly.express as px

from skimage.util import montage
from skimage.filters import rank
from skimage import morphology
from skimage import exposure
from skimage import measure
from skimage import filters
from skimage import io
from skimage import transform
from skimage import registration

from scipy import ndimage
from scipy import signal
from scipy import stats
from scipy import ndimage as ndi


class Eapp():
    def __init__(self, dd_img:np.ndarray, da_img:np.ndarray,
                 ad_img:np.ndarray, aa_img:np.ndarray,
                 abcd_list:list[int,...], G_val:float, **kwargs):
        """ Class for estimating FRET efficiency in image time series.

        __WF_2x_2m instance type as input is recommended__
    
        Parameters
        ----------
        dd_img: ndarray [t,x,y]
            image time series with donor excitation-donor emission
        da_img: ndarray [t,x,y]
            image time series with donor excitation-acceptor emission
        ad_img: ndarray [t,x,y]
            image time series with acceptor excitation-donor emission
        aa_img: ndarray [t,x,y]
            image time series with acceptor excitation-acceptor emission
        abcd_list: list
            list of crosstalk coefficients
        G_val: float
            gauge ("G") parameter of imaging system

        Attributes
        ----------
        DD_img: ndarray [t,x,y]
            image time series with donor excitation-donor emission
            (e.g. 435 nm - CFP ch.)
        DA_img: ndarray [t,x,y]
            image time series with donor excitation-acceptor emission
            (e.g. 435 nm - YFP ch.)
        AD_img: ndarray [t,x,y]
            image time series with acceptor excitation-donor emission
            (e.g. 505 nm - CFP ch.)
        AA_img: ndarray [t,x,y]
            image time series with acceptor excitation-acceptor emission\
            (e.g. 505 nm - YFP ch.)
        a: float
            acceptor bleedthrough coefficient (I_DA(A) / I_AA(A))
        b: float
            acceptor bleedthrough coefficient (I_DD(A) / I_AA(A))
        c: float
            donor bleedthrough coefficient (I_AA(D) / I_DD(D))
        d: float
            donor bleedthrough coefficient (I_DA(D) / I_DD(D))
        G: float
            gauge ("G") parameter of imaging system
        Fc_img: ndarray [t,x,y]
            image time series of sensitized fluorescence
        R_img: ndarray [t,x,y]
            image time series of sensitized fluorescence to donor emission ratio
            (F_c / I_DD)
        Eapp_img: ndarray [t,x,y]
            image time series of E-FRET
        Ecorr_img: ndarray [t,x,y]
            image time series of E-FRET corrected for photobleaching

        """
        self.DD_img = dd_img  # 435-CFP  DD
        self.DA_img = da_img  # 435-YFP  DA
        self.AD_img = ad_img  # 505-CFP  AD
        self.AA_img = aa_img  # 505-YFP  AA

        self.a = abcd_list[0]
        self.b = abcd_list[1]
        self.c = abcd_list[2]
        self.d = abcd_list[3]

        self.G = G_val

        self.Fc_img = self.Fc_calc(dd_img=self.DD_img,
                                   da_img=self.DA_img,
                                   aa_img=self.AA_img,
                                   a=self.a, b=self.b, c=self.c, d=self.d)

        self.R_img, self.Eapp_img = self.E_app_calc(fc_img=self.Fc_img,
                                                    dd_img=self.DD_img,
                                                    G=self.G)
        
        self.Ecorr_img = self.E_cor_calc(e_app_img=self.Eapp_img,
                                         aa_img=self.AA_img,
                                         dd_img=self.DD_img,
                                         c=self.c,
                                         **kwargs)


    @staticmethod
    def Fc_calc(dd_img:np.ndarray, da_img:np.ndarray, aa_img:np.ndarray,
                a:float, b:float, c:float, d:float):
        Fc_img = []
        for frame_num in range(dd_img.shape[0]):
            DD_frame = dd_img[frame_num]
            DA_frame = da_img[frame_num]
            AA_frame = aa_img[frame_num]

            Fc_frame = DA_frame - a*(AA_frame - c*DD_frame) - d*(DD_frame - b*AA_frame)
            Fc_img.append(Fc_frame)

        return np.asarray(Fc_img)


    @staticmethod
    def E_app_calc(fc_img:np.ndarray, dd_img:np.ndarray, G:float):
        R_img = []
        E_app_img = []
        for frame_num in range(fc_img.shape[0]):
            Fc_frame = fc_img[frame_num]
            DD_frame = dd_img[frame_num]

            R_frame = Fc_frame / DD_frame
            R_frame = R_frame
            E_app_frame = R_frame / (R_frame + G)
            E_app_frame[E_app_frame < 0] = 0

            R_img.append(R_frame)
            E_app_img.append(E_app_frame)

        return np.asarray(R_img), np.asarray(E_app_img)


    @staticmethod
    def E_cor_calc(e_app_img:np.ndarray, aa_img:np.ndarray, dd_img:np.ndarray,
                   c:float, mask:np.ndarray, corr_by_mask=False):
        aa_0 = np.mean(aa_img[:2], axis=0)
        dd_0 = np.mean(dd_img[:2], axis=0)

        if corr_by_mask:
            aa_masked = ma.masked_where(~mask, aa_0)
            dd_c_masked = ma.masked_where(~mask, (dd_0*c))
            I_0_val = np.mean(aa_masked - (dd_c_masked * c))
            I_t_series = np.asarray([np.mean(ma.masked_where(~mask, aa_img[i] - (dd_img[i]*c)).compressed()) \
                                     for i in range(aa_img.shape[0])])
        else:
            I_0_val = np.mean(aa_0 - (dd_0 * c))
            I_t_series = np.asarray([np.mean(aa_img[i] - (dd_img[i]*c)) \
                                     for i in range(aa_img.shape[0])])

        E_corr = [e_app_img[i] * (I_0_val / I_t_series[i]) \
                  for i in range(e_app_img.shape[0])]
        
        return np.asarray(E_corr)


# Copyright (c) 2012-2020, Christoph Gohlke
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Read Olympus(r) image files (OIF and OIB).

Copyright (c) 2012-2020, Christoph Gohlke

Oiffile is a Python library to read image and metadata from Olympus Image
Format files. OIF is the native file format of the Olympus FluoView(tm)
software for confocal microscopy.

There are two variants of the format:

* OIF (Olympus Image File) is a multi-file format that includes a main setting
  file (.oif) and an associated directory with data and setting files (.tif,
  .bmp, .txt, .pyt, .roi, and .lut).

* OIB (Olympus Image Binary) is a compound document file, storing OIF and
  associated files within a single file.

__Author__:
  `Christoph Gohlke <https://www.lfd.uci.edu/~gohlke/>`_

__Organization__:
  Laboratory for Fluorescence Dynamics. University of California, Irvine

__License__: BSD 3-Clause

__Version__: 2020.1.18

Requirements
------------
* `CPython >= 3.6 <https://www.python.org>`_
* `Numpy 1.14 <https://www.numpy.org>`_
* `Tifffile 2019.1.1 <https://pypi.org/project/tifffile/>`_

Revisions
---------
2020.1.18
    Fix indentation error.
2020.1.1
    Support multiple image series.
    Parse shape and dtype from settings file.
    Remove support for Python 2.7 and 3.5.
    Update copyright.

Notes
-----
The API is not stable yet and might change between revisions.

No specification document is available.

Tested only with files produced on Olympus FV1000 hardware.

Examples
--------

Read the image from an OIB file as numpy array:

>>> image = OibImread('test.oib')
>>> image.shape
(3, 256, 256)
>>> image[:, 95, 216]
array([820,  50, 436], dtype=uint16)

Read the image from a single TIFF file in an OIB file:

>>> with OifFile('test.oib') as oib:
...     filename = natural_sorted(oib.glob('*.tif'))[0]
...     image = oib.asarray(filename)
>>> filename
'Storage00001/s_C001.tif'
>>> image[95, 216]
820

Access metadata and the OIB main file:

>>> with OifFile('test.oib') as oib:
...     oib.axes
...     oib.shape
...     oib.dtype
...     dataname = oib.mainfile['File Info']['DataName']
'CYX'
(3, 256, 256)
dtype('uint16')
>>> dataname
'Cell 1 mitoEGFP.oib'

Extract the OIB file content to an OIF file and associated data directory:

>>> tempdir = tempfile.mkdtemp()
>>> oib2oif('test.oib', location=tempdir)
Saving ... done.

Read the image from the extracted OIF file:

>>> image = imread(f'{tempdir}/{dataname[:-4]}.oif')
>>> image[:, 95, 216]
array([820,  50, 436], dtype=uint16)

Read OLE compound file and access the 'OibInfo.txt' settings file:

>>> with CompoundFile('test.oib') as com:
...     info = com.open_file('OibInfo.txt')
...     len(com.files())
14
>>> info = SettingsFile(info, 'OibInfo.txt')
>>> info['OibSaveInfo']['Version']
'2.0.0.0'

"""

__version__ = '2020.1.18'

__all__ = (
    'imread',
    'oib2oif',
    'OifFile',
    'OifFileError',
    'OibFileSystem',
    'OifFileSystem',
    'SettingsFile',
    'CompoundFile',
    'filetime',
)

import os
import re
import struct
from io import BytesIO
from glob import glob
from datetime import datetime

import numpy

from tifffile import TiffFile, TiffSequence, lazyattr, natural_sorted


def OibImread(filename, *args, **kwargs):
    """Return image data from OIF or OIB file as numpy array.

    'args' and 'kwargs' are arguments to OifFile.asarray().

    """
    with OifFile(filename) as oif:
        result = oif.asarray(*args, **kwargs)
    return result


def oib2oif(filename, location='', verbose=1):
    """Convert OIB file to OIF."""
    with OibFileSystem(filename) as oib:
        oib.saveas_oif(location=location, verbose=verbose)


class OifFileError(Exception):
    """Exception to raise issues with OIF or OIB structure."""


class OifFile:
    """Olympus Image File.

    Attributes
    ----------
    mainfile : SettingsFile
        The main OIF settings.
    filesystem : OibFileSystem or OifFileSystem
        The underlying file system instance.
    series : tuple of tifffile.TiffSequence
        Sequence of TIFF files. Includes shape, dtype, and axes information.
    is_oib : bool
        True if OIB file.

    """

    def __init__(self, filename):
        """Open OIF or OIB file and read main settings."""
        self.filename = filename
        if filename.lower().endswith('.oif'):
            self.filesystem = OifFileSystem(filename)
            self.is_oib = False
        else:
            self.filesystem = OibFileSystem(filename)
            self.is_oib = True
        self.mainfile = self.filesystem.settings
        # map file names to storage names (flattened name space)
        self._files_flat = {os.path.basename(f): f
                            for f in self.filesystem.files()}

    def open_file(self, filename):
        """Return open file object from path name."""
        try:
            return self.filesystem.open_file(
                self._files_flat.get(filename, filename))
        except (KeyError, OSError) as exc:
            raise FileNotFoundError(f'No such file: {filename}') from exc

    def glob(self, pattern='*'):
        """Return iterator over unsorted file names matching pattern."""
        if pattern == '*':
            return self.filesystem.files()
        pattern = pattern.replace('.', r'\.').replace('*', '.*')
        pattern = re.compile(pattern)
        return (f for f in self.filesystem.files() if pattern.match(f))

    @property
    def axes(self):
        """Return order of axes in image data from mainfile."""
        return self.mainfile['Axis Parameter Common']['AxisOrder'][::-1]

    @property
    def shape(self):
        """Return shape of image data from mainfile."""
        size = {
            self.mainfile[f'Axis {i} Parameters Common']['AxisCode']:
            int(self.mainfile[f'Axis {i} Parameters Common']['MaxSize'])
            for i in range(8)
        }
        return tuple(size[ax] for ax in self.axes)

    @property
    def dtype(self):
        """Return dtype of image data from mainfile."""
        bitcount = int(
            self.mainfile['Reference Image Parameter']['ValidBitCounts'])
        return numpy.dtype('<u2' if bitcount > 8 else '<u2')

    @lazyattr
    def series(self):
        """Return tuple of TiffSequence of sorted TIFF files."""
        series = {}
        for fname in self.glob('*.tif'):
            key = ''.join(c for c in os.path.split(fname)[-1][:-4]
                          if c.isalpha())
            if key in series:
                series[key].append(fname)
            else:
                series[key] = [fname]
        series = [TiffSequence(natural_sorted(files),
                               imread=self.asarray, pattern='axes')
                  for files in series.values()]
        if len(series) > 1:
            series = tuple(reversed(sorted(series,
                                           key=lambda x: len(x.files))))
        return series

    @property
    def tiffs(self):
        """Return first TiffSequence."""
        # required for compatibility with cmapfile < 2020.1.1
        return self.series[0]

    def asarray(self, series=0, **kwargs):
        """Return image data from TIFF file(s) as numpy array.

        By default the data from the TIFF files in the first image series
        is returned.

        The kwargs parameters are passed to the asarray functions of the
        TiffFile or TiffSequence instances.

        """
        if isinstance(series, int):
            return self.series[series].asarray(**kwargs)
        with TiffFile(self.open_file(series), name=series) as tif:
            result = tif.asarray(**kwargs)
        return result

    def close(self):
        """Close file handle."""
        self.filesystem.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __str__(self):
        """Return string with information about OifFile."""
        # info = self.mainfile['Version Info']
        s = [
            self.__class__.__name__,
            os.path.normpath(os.path.normcase(self.filename)),
            f'axes: {self.axes}',
            'shape: {}'.format(', '.join(str(i) for i in self.shape)),
            f'dtype: {self.dtype}',
            # f'system name: {info.get("SystemName", "None")}',
            # f'system version: {info.get("SystemVersion", "None")}',
            # f'file version: {info.get("FileVersion", "None")}',
        ]
        if len(self.series) > 1:
            s.append(f'series: {len(self.series)}')
        return '\n '.join(s)


class OifFileSystem:
    """Olympus Image File file system."""

    def __init__(self, filename, storage_ext='.files'):
        """Open OIF file and read settings."""
        self.filename = filename
        self._path, self.mainfile = os.path.split(os.path.abspath(filename))
        self.settings = SettingsFile(filename, name=self.mainfile)
        self.name = self.settings['ProfileSaveInfo']['Name']
        self.version = self.settings['ProfileSaveInfo']['Version']
        # check that storage directory exists
        storage = os.path.join(self._path, self.mainfile + storage_ext)
        if not os.path.exists(storage) or not os.path.isdir(storage):
            raise OSError(
                f'OIF storage path not found: {self.mainfile}{storage_ext}')
        # list all files
        pathlen = len(self._path + os.path.sep)
        self._files = [self.mainfile]
        for f in glob(os.path.join(storage, '*')):
            self._files.append(f[pathlen:])

    def open_file(self, filename):
        """Return file object from path name."""
        return open(os.path.join(self._path, filename), 'rb')

    def files(self):
        """Return iterator over unsorted files in OIF."""
        return iter(self._files)

    def glob(self, pattern):
        """Return iterator of path names matching the specified pattern."""
        pattern = pattern.replace('.', r'\.').replace('*', '.*')
        pattern = re.compile(pattern)
        return (f for f in self.files() if pattern.match(f))

    def close(self):
        """Close file handle."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __str__(self):
        """Return string with information about OifFileSystem."""
        return '\n '.join((
            self.__class__.__name__,
            os.path.normpath(os.path.normcase(self.filename)),
            f'name: {self.name}',
            f'version: {self.version}',
            f'mainfile: {self.mainfile}',
        ))


class OibFileSystem:
    """Olympus Image Binary file system."""

    def __init__(self, filename):
        """Open compound document and read OibInfo.txt settings."""
        self.filename = filename
        self.com = CompoundFile(filename)
        info = SettingsFile(self.com.open_file('OibInfo.txt'),
                            'OibInfo.txt')['OibSaveInfo']
        self.name = info.get('Name', None)
        self.version = info.get('Version', None)
        self.compression = info.get('Compression', None)
        self.mainfile = info[info['MainFileName']]
        # map OIB file names to CompoundFile file names
        oibfiles = {os.path.split(i)[-1]: i for i in self.com.files()}
        self._files = {v: oibfiles[k] for k, v in info.items()
                       if k.startswith('Stream')}
        # map storage names to directory names
        self._folders = {i[0]: i[1] for i in info.items()
                         if i[0].startswith('Storage')}
        # read main settings file
        self.settings = SettingsFile(self.open_file(self.mainfile),
                                     name=self.mainfile)

    def open_file(self, filename):
        """Return file object from case sensitive path name."""
        try:
            return self.com.open_file(self._files[filename])
        except KeyError as exc:
            raise FileNotFoundError(f'No such file: {filename}') from exc

    def files(self):
        """Return iterator over unsorted files in OIB."""
        return iter(self._files.keys())

    def saveas_oif(self, location='', verbose=0):
        """Save all streams in OIB file as separate files.

        Raise OSError if target files or directories already exist.

        The main .oif file name and storage names are determined from the
        OibInfo.txt settings.

        """
        if location and not os.path.exists(location):
            os.makedirs(location)
        mainfile = os.path.join(location, self.mainfile)
        if os.path.exists(mainfile):
            raise FileExistsError(mainfile + ' already exists')
        for folder in self._folders.keys():
            folder = os.path.join(location, self._folders.get(folder, ''))
            if os.path.exists(folder):
                raise FileExistsError(folder + ' already exists')
            os.makedirs(folder)
        if verbose:
            print('Saving', mainfile, end=' ')
        for f in self._files.keys():
            folder, name = os.path.split(f)
            folder = os.path.join(location, self._folders.get(folder, ''))
            path = os.path.join(folder, name)
            if verbose == 1:
                print(end='.')
            elif verbose > 1:
                print(path)
            with open(path, 'w+b') as fh:
                fh.write(self.open_file(f).read())
        if verbose == 1:
            print(' done.')

    def close(self):
        """Close file handle."""
        self.com.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __str__(self):
        """Return string with information about OibFileSystem."""
        return '\n '.join((
            self.__class__.__name__,
            os.path.normpath(os.path.normcase(self.filename)),
            f'name: {self.name}',
            f'version: {self.version}',
            f'mainfile: {self.mainfile}',
            f'compression: {self.compression}',
        ))


class SettingsFile(dict):
    """Olympus settings file (oif, txt, pyt, roi, lut).

    Settings files contain little endian utf-16 encoded strings, except for
    [ColorLUTData] sections, which contain uint8 binary arrays.

    Settings can be accessed as a nested dictionary {section: {key: value}},
    except for {'ColorLUTData': numpy array}.

    """

    def __init__(self, arg, name=None):
        """Read settings file and parse into nested dictionaries.

        Parameters
        ----------
        arg : str or file object
            Name of file or open file containing little endian UTF-16 string.
            File objects are closed by this function.
        name : str
            Human readable label of stream.

        """
        dict.__init__(self)
        if isinstance(arg, str):
            self.name = arg
            stream = open(arg, 'rb')
        else:
            self.name = str(name)
            stream = arg

        try:
            content = stream.read()
        finally:
            stream.close()

        if content[:4] == b'\xFF\xFE\x5B\x00':
            # UTF16 BOM
            content = content.rsplit(
                b'[\x00C\x00o\x00l\x00o\x00r\x00L\x00U\x00T\x00'
                b'D\x00a\x00t\x00a\x00]\x00\x0D\x00\x0A\x00', 1)
            if len(content) > 1:
                self['ColorLUTData'] = numpy.fromstring(
                    content[1], 'uint8').reshape(-1, 4)
            content = content[0].decode('utf-16')
        elif content[:1] == b'[':
            # try UTF-8
            content = content.rsplit(b'[ColorLUTData]\r\n', 1)
            if len(content) > 1:
                self['ColorLUTData'] = numpy.fromstring(
                    content[1], 'uint8').reshape(-1, 4)
            try:
                content = content[0].decode('utf-8')
            except Exception as exc:
                raise ValueError('not a valid settings file') from exc
        else:
            raise ValueError('not a valid settings file')

        for line in content.splitlines():
            line = line.strip()
            if line.startswith(';'):
                continue
            if line.startswith('[') and line.endswith(']'):
                self[line[1:-1]] = properties = {}
            else:
                key, value = line.split('=')
                properties[key] = astype(value)

    def __str__(self):
        """Return string with information about SettingsFile."""
        return '\n '.join((self.__class__.__name__, format_dict(self)))


class CompoundFile:
    """Compound Document File.

    A partial implementation of the "[MS-CFB] - v20120705, Compound File
    Binary File Format" specification by Microsoft Corporation.

    This should be able to read Olympus OIB and Zeiss ZVI files.

    """

    MAXREGSECT = 0xFFFFFFFA
    DIFSECT = 0xFFFFFFFC
    FATSECT = 0xFFFFFFFD
    ENDOFCHAIN = 0xFFFFFFFE
    FREESECT = 0xFFFFFFFF
    MAXREGSID = 0xFFFFFFFA
    NOSTREAM = 0xFFFFFFFF

    def __init__(self, filename):
        """Initialize instance from file."""
        self._fh = open(filename, 'rb')
        if self._fh.read(8) != b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1':
            self.close()
            raise ValueError('not a compound document file')
        (
            self.clsid,
            self.version_minor,
            self.version_major,
            byteorder,
            sector_shift,
            mini_sector_shift,
            _,
            _,
            self.dir_len,
            self.fat_len,
            self.dir_start,
            _,
            self.mini_stream_cutof_size,
            self.minifat_start,
            self.minifat_len,
            self.difat_start,
            self.difat_len,
        ) = struct.unpack('<16sHHHHHHIIIIIIIIII', self._fh.read(68))

        self.byteorder = {0xFFFE: '<', 0xFEFF: '>'}[byteorder]
        if self.byteorder != '<':
            raise NotImplementedError('big-endian byte order not supported')

        if self.clsid == b'\x00' * 16:
            self.clsid = None
        if self.clsid is not None:
            raise OifFileError(f'cannot handle clsid {self.clsid}')

        if self.version_minor != 0x3E:
            raise OifFileError(
                f'cannot handle version_minor {self.version_minor}')
        if mini_sector_shift != 0x0006:
            raise OifFileError(
                f'cannot handle mini_sector_shift {mini_sector_shift}')
        if not (
            (self.version_major == 0x4 and sector_shift == 0x000C) or
            (self.version_major == 0x3 and sector_shift == 0x0009 and
             self.dir_len == 0)
        ):
            raise OifFileError(
                f'cannot handle version_major {self.version_major} '
                f'and sector_shift {sector_shift}'
            )

        self.filename = filename
        self.sec_size = 2**sector_shift
        self.short_sec_size = 2**mini_sector_shift

        secfmt = '<' + ('I' * (self.sec_size // 4))
        # read DIFAT
        self._difat = list(
            struct.unpack('<' + ('I' * 109), self._fh.read(436))
        )
        nextsec = self.difat_start
        for i in range(self.difat_len):
            if nextsec >= CompoundFile.MAXREGSID:
                raise OifFileError('nextsec >= CompoundFile.MAXREGSID')
            sec = struct.unpack(secfmt, self._sec_read(nextsec))
            self._difat.extend(sec[:-1])
            nextsec = sec[-1]
        # if nextsec != CompoundFile.ENDOFCHAIN: raise OifFileError()
        self._difat = self._difat[:self.fat_len]
        # read FAT
        self._fat = []
        for secid in self._difat:
            self._fat.extend(struct.unpack(secfmt, self._sec_read(secid)))
        # read mini FAT
        self._minifat = []
        for i, sector in enumerate(self._sec_chain(self.minifat_start)):
            if i >= self.minifat_len:
                break
            self._minifat.extend(struct.unpack(secfmt, sector))
        # read directories
        self._dirs = []
        for sector in self._sec_chain(self.dir_start):
            for i in range(0, self.sec_size, 128):
                direntry = DirectoryEntry(sector[i:i+128], self.version_major)
                self._dirs.append(direntry)
        # read root storage
        if len(self._dirs) <= 0:
            raise OifFileError('no directories found')
        root = self._dirs[0]
        if root.name != 'Root Entry':
            raise OifFileError(
                f"no root directory found, got {root.name}")
        if root.create_time is not None:  # and root.modify_time is None
            raise OifFileError(f'invalid root.create_time {root.create_time}')
        if root.stream_size % self.short_sec_size != 0:
            raise OifFileError(
                f'root.stream_size {root.stream_size} does not match '
                f'short_sec_size {self.short_sec_size}'
            )
        # read mini stream
        self._ministream = b''.join(self._sec_chain(root.sector_start))
        self._ministream = self._ministream[:root.stream_size]
        # map stream/file names to directory entries
        nostream = CompoundFile.NOSTREAM
        join = '/'.join  # os.path.sep.join
        dirs = self._dirs
        visited = [False] * len(self._dirs)

        def parse(dirid, path):
            # return iterator over all file names and their directory entries
            # TODO: replace with red-black tree parser
            if dirid == nostream or visited[dirid]:
                return
            visited[dirid] = True
            de = dirs[dirid]
            if de.is_stream:
                yield join(path + [de.name]), de
            yield from parse(de.left_sibling_id, path)
            yield from parse(de.right_sibling_id, path)
            if de.is_storage:
                yield from parse(de.child_id, path + [de.name])

        self._files = dict(parse(self._dirs[0].child_id, []))

    def _read_stream(self, direntry):
        """Return content of stream."""
        if direntry.stream_size < self.mini_stream_cutof_size:
            result = b''.join(self._mini_sec_chain(direntry.sector_start))
        else:
            result = b''.join(self._sec_chain(direntry.sector_start))
        return result[:direntry.stream_size]

    def _sec_read(self, secid):
        """Return content of sector from file."""
        self._fh.seek(self.sec_size + secid * self.sec_size)
        return self._fh.read(self.sec_size)

    def _sec_chain(self, secid):
        """Return iterator over FAT sector chain content."""
        while secid != CompoundFile.ENDOFCHAIN:
            if secid <= CompoundFile.MAXREGSECT:
                yield self._sec_read(secid)
            secid = self._fat[secid]

    def _mini_sec_read(self, secid):
        """Return content of sector from mini stream."""
        pos = secid * self.short_sec_size
        return self._ministream[pos: pos+self.short_sec_size]

    def _mini_sec_chain(self, secid):
        """Return iterator over mini FAT sector chain content."""
        while secid != CompoundFile.ENDOFCHAIN:
            if secid <= CompoundFile.MAXREGSECT:
                yield self._mini_sec_read(secid)
            secid = self._minifat[secid]

    def files(self):
        """Return sequence of file names."""
        return self._files.keys()

    def direntry(self, name):
        """Return DirectoryEntry of filename."""
        return self._files[name]

    def open_file(self, filename):
        """Return stream as file like object."""
        return BytesIO(self._read_stream(self._files[filename]))

    def format_tree(self):
        """Return formatted string with list of all files."""
        return '\n'.join(natural_sorted(self.files()))

    def close(self):
        """Close file handle."""
        self._fh.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __str__(self):
        """Return string with information about CompoundFile."""
        s = [
            self.__class__.__name__,
            os.path.normpath(os.path.normcase(self.filename)),
        ]
        for attr in (
            'clsid', 'version_minor', 'version_major', 'byteorder',
            'dir_len', 'fat_len', 'dir_start', 'mini_stream_cutof_size',
            'minifat_start', 'minifat_len', 'difat_start', 'difat_len'
        ):
            s.append(f'{attr}: {getattr(self, attr)}')
        return '\n '.join(s)


class DirectoryEntry:
    """Compound Document Directory Entry."""

    __slots__ = (
        'name', 'entry_type', 'color', 'left_sibling_id', 'right_sibling_id',
        'child_id', 'clsid', 'user_flags', 'create_time', 'modify_time',
        'sector_start', 'stream_size', 'is_stream', 'is_storage'
    )

    def __init__(self, data, version_major):
        """Initialize directory entry from 128 bytes."""
        (
            name,
            name_len,
            self.entry_type,
            self.color,
            self.left_sibling_id,
            self.right_sibling_id,
            self.child_id,
            self.clsid,
            self.user_flags,
            create_time,
            modify_time,
            self.sector_start,
            self.stream_size,
        ) = struct.unpack('<64sHBBIII16sIQQIQ', data)

        if version_major == 3:
            self.stream_size = struct.unpack('<I', data[-8:-4])[0]
        if self.clsid == b'\000' * 16:
            self.clsid = None

        if name_len % 2 != 0 or name_len > 64:
            raise OifFileError(f'invalid name_len {name_len}')
        if self.color not in (0, 1):
            raise OifFileError(f'invalid color {self.color}')

        self.name = name[:name_len-2].decode('utf-16')
        self.create_time = filetime(create_time)
        self.modify_time = filetime(modify_time)
        self.is_stream = self.entry_type == 2
        self.is_storage = self.entry_type == 1

    def __str__(self):
        """Return string with information about DirectoryEntry."""
        s = [self.__class__.__name__]
        for attr in self.__slots__[1:]:
            s.append(f'{attr}: {getattr(self, attr)}')
        return '\n '.join(s)


def format_dict(adict, prefix=' ', indent=' ', bullets=('', ''),
                excludes=('_', ), linelen=79, trim=1):
    """Return pretty-print of nested dictionary."""
    result = []
    for k, v in sorted(adict.items(), key=lambda x: str(x[0]).lower()):
        if any(k.startswith(e) for e in excludes):
            continue
        if isinstance(v, dict):
            v = '\n' + format_dict(v, prefix=prefix+indent, excludes=excludes,
                                   trim=0)
            result.append(f'{prefix}{bullets[1]}{k}: {v}')
        else:
            result.append((f'{prefix}{bullets[0]}{k}: {v}')[:linelen].rstrip())
    if trim > 0:
        result[0] = result[0][trim:]
    return '\n'.join(result)


def astype(value, types=None):
    """Return argument as one of types if possible."""
    if value[0] in '\'"':
        return value[1:-1]
    if types is None:
        types = int, float, str
    for typ in types:
        try:
            return typ(value)
        except (ValueError, TypeError, UnicodeEncodeError):
            pass
    return value


def filetime(ft):
    """Return Python datetime from Microsoft FILETIME number."""
    if not ft:
        return None
    sec, nsec = divmod(ft - 116444736000000000, 10000000)
    return datetime.utcfromtimestamp(sec).replace(microsecond=(nsec // 10))


if __name__ == '__main__':
    import doctest
    import tempfile  # noqa: used in doctrings

    numpy.set_printoptions(suppress=True, precision=5)
    doctest.testmod(optionflags=doctest.ELLIPSIS)


import numpy as np
from numpy import ma

import matplotlib.pyplot as plt
import matplotlib.patches as patches

from skimage import morphology
from skimage import exposure
from skimage import measure
from skimage import filters
from skimage import segmentation

from scipy import ndimage as ndi
from scipy import stats

from ..utils import masking
from ..utils import plot


class wt_vs_mut_multistim():
    def __init__(self, wt_img:np.ndarray, mut_img:np.ndarray,
                 proc_mask:np.ndarray, narrow_proc_mask:np.ndarray,
                 **kwargs):
        """ Class is designed to create differential images and insertion masks
        for two separate image time series (for instance, wild-type NCS and mutant NCS,
        or two distinct NCSs). The up_mask_calc method is responsible for detecting
        insertion regions.

        Suitable for image series with __multiple__ application/stimulation only.

        The class attributes contain various types of insertion masks
        and corresponding profiles for these masks (raw intensity and ΔF/F profiles).

        __Requires wf_x2_m2 instance type as input!__

        Parameters
        ----------
        wt_img: ndarray [t,x,y]
            wild type images time series
        mut_img: ndarray [t,x,u]
            mutant images time series
        proc_mask: ndarray [x,y]
            cell processes boolean mask, extended
        narrow_proc_mask: ndarray [x,y]
            cell processes boolean mask, unextended

        Attributes
        ----------
        wt_img: ndarray [t,x,y]
            wild type images time series
        mut_img: ndarray [t,x,y]
            mutant images time series
        proc_mask: ndarray [x,y]
            cell processes boolean mask, extended
        narrow_proc_mask: ndarray [x,y]
            cell processes boolean mask, unextended
        wt_up_mask_list: ndarray [stimuli, x,y]
            boolean mask of intensity increase regions for wild-type channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        wt_up_label_list: ndarray [stimuli, x,y]
            label image of intensity increase regions for wild-type channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        wt_diff_img_list: ndarray [stimuli, x,y]
            differential image of intensity changes after stimulation for wild-type channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        mut_up_mask_list: ndarray [stimuli, x,y]
            boolean mask of intensity increase regions for mutant/2nd NCS channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        mut_up_label_list: ndarray [stimuli, x,y]
            label image of intensity increase regions for mutant/2nd NCS channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        mut_diff_img_list: ndarray [stimuli, x,y]
            differential image of intensity changes after stimulation for mutant/2nd NCS channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        connected_up_mask_list: ndarray [stimuli, x,y]
            regions of `wt_up_mask` which overlay with `mut_up_mask` (pairwise for each stimulation),
            created with `utils.masking.mask_connection()`

        """
        self.wt_img = wt_img
        self.mut_img = mut_img
        self.proc_mask = proc_mask
        self.narrow_proc_mask = narrow_proc_mask

        cell_contour_mask = np.copy(self.proc_mask)
        cell_contour_mask[self.narrow_proc_mask] = 0
        # WT masking
        self.wt_up_mask_list, self.wt_up_label_list, self.wt_diff_img_list = self.up_mask_calc(self.wt_img,
                                                                                               cell_contour_mask,
                                                                                               **kwargs)
        
        # mutant masking
        self.mut_up_mask_list, self.mut_up_label_list, self.mut_diff_img_list = self.up_mask_calc(self.mut_img,
                                                                                                  cell_contour_mask,
                                                                                                  **kwargs)

        # self.wt_diff_img_series = self.diff_img_series_calc(self.wt_img)
        # self.mut_diff_img_series = self.diff_img_series_calc(self.mut_img)

        # masks modification
        self.connected_up_mask_list, self.connected_up_label_list = np.empty_like(self.proc_mask), np.empty_like(self.proc_mask)
        for stim_num in range(self.wt_up_mask_list.shape[0]):
            connected_up_mask, connected_up_label = masking.mask_connection(input_master_mask=self.wt_up_mask_list[stim_num],
                                                                            input_minor_mask=self.mut_up_mask_list[stim_num])
            self.connected_up_mask_list = np.dstack((self.connected_up_mask_list,
                                                     connected_up_mask))
            self.connected_up_label_list = np.dstack((self.connected_up_label_list,
                                                     connected_up_label))
        self.connected_up_mask_list = np.moveaxis(self.connected_up_mask_list, -1, 0)
        self.connected_up_label_list = np.moveaxis(self.connected_up_label_list, -1, 0)


    @staticmethod
    def up_mask_calc(input_img_series:np.ndarray, input_img_mask:np.ndarray,
                     sd_tolerance:int=2, stim_list:list[int]=[10],
                     base_frames:int=2, stim_win:int=1, stim_shift:int=0):
        """ Function for generating a set of insertion regions mask
        for image series with multiple stimuli/applications using the differential image method.

        Could be used in stand-alone mode as a static method of the WTvsMut class.

        Parameters
        ----------
        input_img_series: ndarray [t,x,y]
            image time series
        input_img_mask: ndarray [x,y]
            cell region boolean mask
        sd_tolerance: int
            insertion ("up") region detection threshold: the number
            of standard deviations of extracellular noise
            (measured in the area outside of `input_img_mask`)
        base_frames: int
            number of frames from the beginning
            of the image series used to create
            an image of basal fluorescence
        stim_list: list[int]
            indexes of the frames where stimulations begins
        stim_win: int
            stimulation window, the number of frames following each `stim_list` indexes
            that are used to create set of images displaying maximal insertions after each stimuli

        Returns
        -------
        up_mask_filt_list: ndarray [stimuli, x,y]
            boolean masks of intensity increase regions
        up_label_list: ndarray [stimuli, x,y]
            label images of intensity increase regions  
        img_diff_list: ndarray [stimuli, x,y]
            differential images of intensity changes after stimulation

        """
        ref_img_series = filters.gaussian(input_img_series, sigma=1.5, channel_axis=0)

        up_mask_filt_list = []
        up_label_list = []
        img_diff_list = []
        stim_i_pairs_list = []
        for stim_i in stim_list:
            base_i = stim_i-base_frames

            img_base = np.mean(ref_img_series[base_i:stim_i], axis=0) 

            start_i = stim_i+stim_shift
            fin_i = stim_i+stim_win + stim_shift
            img_max = np.mean(ref_img_series[start_i:fin_i], axis=0)

            img_diff = img_max - img_base
            img_diff_norm = img_diff / np.max(np.abs(img_diff))

            diff_sd = np.std(img_diff, where=input_img_mask)
            up_mask = img_diff > diff_sd * sd_tolerance
            # up_mask =  img_diff_norm > sd_tolerance

            # NEW MASK PROCESSING
            up_mask_filt = ndi.binary_fill_holes(up_mask)
            up_mask_filt = morphology.remove_small_objects(up_mask_filt, min_size=20)
            up_mask_filt = segmentation.expand_labels(up_mask_filt, distance=1)

            # OLD MASK PROCESSING
            # up_mask_filt = morphology.opening(up_mask, footprint=morphology.disk(3))
            # up_mask_filt = morphology.dilation(up_mask_filt, footprint=morphology.disk(2))

            up_mask_filt = segmentation.clear_border(up_mask_filt)
            up_label = measure.label(up_mask_filt)

            up_mask_filt_list.append(up_mask_filt)
            up_label_list.append(up_label)
            img_diff_list.append(img_diff_norm)

            stim_i_pairs_list.append([[base_i, stim_i],[start_i, fin_i]])

        print(stim_i_pairs_list)

        return np.asarray(up_mask_filt_list), np.asarray(up_label_list), np.asarray(img_diff_list)


    @staticmethod
    def up_mask_calc_old(input_img_series:np.ndarray, input_img_mask:np.ndarray,
                     sd_tolerance:int=2, base_frames:int=5,
                     stim_list:list[int]=[10], stim_win:int=5):
        """ Function for generating a set of insertion regions mask
        for image series with multiple stimuli/applications using the differential image method.

        Could be used in stand-alone mode as a static method of the WTvsMut class.

        Parameters
        ----------
        input_img_series: ndarray [t,x,y]
            image time series
        input_img_mask: ndarray [x,y]
            cell region boolean mask
        sd_tolerance: int
            insertion ("up") region detection threshold: the number
            of standard deviations of extracellular noise
            (measured in the area outside of `input_img_mask`)
        base_frames: int
            number of frames from the beginning
            of the image series used to create
            an image of basal fluorescence
        stim_list: list[int]
            indexes of the frames where stimulations begins
        stim_win: int
            stimulation window, the number of frames following each `stim_list` indexes
            that are used to create set of images displaying maximal insertions after each stimuli

        Returns
        -------
        up_mask_filt_list: ndarray [stimuli, x,y]
            boolean masks of intensity increase regions
        up_label_list: ndarray [stimuli, x,y]
            label images of intensity increase regions  
        img_diff_list: ndarray [stimuli, x,y]
            differential images of intensity changes after stimulation

        """
        ref_img_series = filters.gaussian(input_img_series, sigma=1.25, channel_axis=0)

        img_base = np.mean(ref_img_series[:base_frames+1], axis=0)

        up_mask_filt_list = []
        up_label_list = []
        img_diff_list = []
        stim_i_pairs_list = []
        for stim_start in stim_list:
            start_i = stim_start
            stop_i = stim_start+stim_win+1

            stim_i_pairs_list.append([start_i, stop_i-1])

            img_max = np.mean(ref_img_series[start_i:stop_i], axis=0)

            img_diff = img_max - img_base
            img_diff_norm = img_diff/np.max(np.abs(img_diff))

            diff_sd = np.std(img_diff, where=input_img_mask)
            up_mask = img_diff > diff_sd * sd_tolerance
            # up_mask =  img_diff_norm > sd_tolerance

            up_mask_filt = morphology.opening(up_mask, footprint=morphology.disk(2))
            up_mask_filt = morphology.dilation(up_mask_filt, footprint=morphology.disk(1))
            up_mask_filt = segmentation.clear_border(up_mask_filt)
            up_label = measure.label(up_mask_filt)

            up_mask_filt_list.append(up_mask_filt)
            up_label_list.append(up_label)
            img_diff_list.append(img_diff_norm)

        print(stim_i_pairs_list)

        return np.asarray(up_mask_filt_list), np.asarray(up_label_list), np.asarray(img_diff_list)


    @staticmethod
    def diff_img_series_calc(input_img_series:np.ndarray, diff_step=1):
        diff_img_series = []
        for i in range(input_img_series.shape[0]-diff_step):
            img_1 = input_img_series[i]
            img_2 = input_img_series[i+diff_step]
            diff_img = img_2 - img_1
            diff_img_series.append(diff_img)
        return np.asarray(diff_img_series)


    def diff_img_pic(self):
        """ Differential images for each simulation.
        
        """
        cell_contour = measure.find_contours(self.proc_mask, level=0.5)

        stim_num = self.wt_up_mask_list.shape[0]
        fig = plt.figure(figsize= (30, 30))
        for i in range(stim_num):
            ax0 = fig.add_subplot(2, stim_num, i+1)
            ax0.text(40, 40, f'WT ({int(np.max(self.wt_up_label_list[i]))} ROIs)', fontsize=20, color='white')
            ax0.imshow(self.wt_diff_img_list[i], cmap=plot.CMaps().cmap_red_green, vmax=1, vmin=-1)
            for ce_c in cell_contour:
                ax0.plot(ce_c[:, 1], ce_c[:, 0], linewidth=1, color='w')
            ax0.axis('off')
            ax0.axis('off')

            ax1 = fig.add_subplot(2, stim_num, i+1+stim_num)
            ax1.text(40, 40, f'Mut. ({int(np.max(self.mut_up_label_list[i]))} ROIs)', fontsize=20, color='white')
            ax1.imshow(self.mut_diff_img_list[i], cmap=plot.CMaps().cmap_red_green, vmax=1, vmin=-1)
            for ce_c in cell_contour:
                ax1.plot(ce_c[:, 1], ce_c[:, 0], linewidth=2, color='w')
            ax1.axis('off')
        plt.tight_layout()
        plt.show()


    def mask_diff_pic(self):
        """ WT vs. mutant up masks comparison with overlay (red - WT, green - mutant).
        
        """
        cell_contour = measure.find_contours(self.proc_mask, level=0.5)

        stim_num = self.wt_up_mask_list.shape[0]
        fig = plt.figure(figsize= (20, 20))
        for i in range(stim_num):
            wt_mask = np.asarray(self.wt_up_mask_list[i], dtype=int) 
            mut_mask = np.asarray(self.mut_up_mask_list[i], dtype=int)
            sum_mask = wt_mask + mut_mask
            tot_mask = sum_mask != 0

            # up_mask_diff = plot.toRGB(r_img=wt_mask,
            #                           g_img=mut_mask,
            #                           b_img=np.zeros_like(wt_mask))
            ax0 = plt.subplot(1, stim_num, i+1)
            ax0.imshow(np.mean(self.mut_img, axis=0)*-1, cmap='Greys')
            ax0.imshow(ma.masked_where(~tot_mask, sum_mask)*-1, cmap=plot.CMaps().cmap_red_green)
            ax0.text(100, 100, f'WT {int(np.max(self.wt_up_label_list[i]))} ROIs\nMut. {int(np.max(self.mut_up_label_list[i]))} ROIs',
                     fontsize=20, color='white')
            for ce_c in cell_contour:
                ax0.plot(ce_c[:, 1], ce_c[:, 0], linewidth=2, color='w')
            ax0.axis('off')

        plt.tight_layout()
        plt.show()


import time, os, pickle
import autograd.numpy as np
from pathos.helpers import cpu_count
from pathos.multiprocessing import ProcessingPool
from utils import kalman_filter, RTS_smoother, non_Gauss_1D_filtering, \
				  non_Gauss_1D_smoothing, diag_gaussian_2d_pdf
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
		

class state_est():

	def __init__(self, analyze_FRET_data_inst, param_est_inst):
		"""
		Constructor takes `analyze_FRET_data' instance and 'param_est' 
		instance and inherits their attributes. Parameter estimation
		is run prior to this to generate a sample of parameters
		from the parameter posterior distribution.
		"""
		
		self.a = analyze_FRET_data_inst
		self.p = param_est_inst
		self.ssm = self.p.ssm
		self.ps = self.p.post_p_samples
		
		out_dir = self.a.res_dir + '/state_dists'
		if not os.path.exists(out_dir):
			os.makedirs(out_dir)
		
	def single_p_integrate_Gaussian(self, p):
		"""
		Get the state posterior distribution for Gaussian process noise, which
		is solved in closed form using the Kalman filter and RTS smoother.
		This is done for a fixed parameter vector `p'.
		"""
		
		# Forward pass / Kalman filter
		m_pred, P_pred, m, P, _, _ = kalman_filter(self.ssm.Q(p), 
												   self.ssm.H(p), 
												   self.ssm.R(), 
												   self.ssm.y(), 
												   self.ssm.m_0, 
												   self.ssm.P_0, 
												   self.a.tDD)
	
		# Backwards pass / Rauch-Tung-Streibel smoother
		ms, Ps = RTS_smoother(m_pred, P_pred, m, P)
		
		return [ms, Ps]
	
	def single_p_integrate_non_Gaussian(self, p):
		"""
		Get the state posterior distribution for non-Gaussian process noise.
		This is done by directly integrating the Bayesian filtering and 
		smoothing equations, which is possible since the state (chi) is only
		1-dimensional. This is done for a fixed parameter vector `p'.
		"""
		
		# Define state transition pdf; here we assume deterministic
		# dynamics are trivial, x --> x + q; q is defined by q_tilde
		idx_lo = 2 + self.a.num_params_f_D + self.a.num_params_f_A
		idx_hi = idx_lo + self.a.num_params_Q_tilde
		Q = self.a.m.Q_tilde.pdf(p[idx_lo:idx_hi])
		
		m_0 = self.a.anl_params.m0_chi
		P_0 = self.a.anl_params.P0_chi
		
		# This is defined in the param_est object
		chis = self.p.calc_non_gauss_chi_range()
	
		# Observation likelihoods
		LHs = self.p.calc_obs_likelihoods(p, chis)
		
		# Filter and smooth by integrating over chi at each timestep
		dist = non_Gauss_1D_filtering(chis, Q, LHs, m_0, P_0, self.a.tDD)
		dist = non_Gauss_1D_smoothing(dist)
								  
		return dist
		
	def single_p_SMC_forward(self, p):
		"""
		Get the state posterior distribution for non-Gaussian process noise.
		This is done by using a particle filter. This only gets the filtering
		distribution by propagaint particles from t=0 to t=end. This is done
		for a fixed parameter vector `p'.
		"""
		
		N = self.a.anl_params.num_SMC_samples
		
		# Pick out parameters for Q_tilde distribution
		idx_lo = 2 + self.a.num_params_f_D + self.a.num_params_f_A
		idx_hi = idx_lo + self.a.num_params_Q_tilde
		Q_chi = self.a.m.Q_tilde.sample(p[idx_lo:idx_hi],
										size=(self.a.num_IDD, N))
		
		# Initialize particles from chi prior and give uniform weights
		xs = np.zeros((self.a.num_IDD, 2, N), dtype=np.float16)
		xs[0] = np.random.multivariate_normal(self.ssm.m_0, self.ssm.P_0,
											  size=N).T
		var1 = self.ssm.R()[:, 0, 0]
		var2 = self.ssm.R()[:, 1, 1]
		y = self.ssm.y()
		H = self.ssm.H(p)
		
		for i in range(self.a.num_IDD - 1):
			
			# Particle loctions
			const_pred = xs[i, 0]
			chi_pred = xs[i, 1] + Q_chi[i]
			x_pred = np.vstack((const_pred.T, chi_pred.T))
			
			# Bootstrap filter; don't need normalized prob w/ resampling
			res1 = y[i, 0] - (H[i, 0, 0]*const_pred + H[i, 0, 1]*chi_pred)
			res2 = y[i, 1] - (H[i, 1, 0]*const_pred + H[i, 1, 1]*chi_pred)
			prob  = np.exp(-res1**2/2/var1[i] - res2**2/2/var2[i])
			ws = prob/np.sum(prob)
			idxs = np.random.choice(range(N), N, p=ws)
			xs[i + 1] = x_pred[:, idxs]
			
		return xs
	
	def single_p_SMC_backward(self, p, xs):
		"""
		Get the state posterior smoothing distribution for non-Gaussian 
		process noise, using the xs and weights from the forward-pass 
		particle filter. 
		"""
		
		N = self.a.anl_params.num_SMC_samples
		
		# ws shape  is timesteps, num_particles
		ws = np.zeros((self.a.num_IDD, N), dtype=np.float16)
		ws[-1] = np.ones(N)/N
		
		# Pick out parameters for Q_tilde distribution
		idx_lo = 2 + self.a.num_params_f_D + self.a.num_params_f_A
		idx_hi = idx_lo + self.a.num_params_Q_tilde
		
		for i in range(self.a.num_IDD - 2, -1, -1):
			
			# Row is j = t + 1 index, column is t index
			mat_diffs = xs[i + 1, 1][:, None] - xs[i, 1][None, :]
			probs = self.a.m.Q_tilde.pdf(p[idx_lo:idx_hi])(mat_diffs)
			
			# Sum_k p(x_t+1(j) | x_t(k))
			den = np.sum(probs, axis=-1)
			
			# Sum_j ws_t+1(j) p(x_t+1(j) | x_t(i))
			# Take transponse of probs since sum is over 2nd index
			w = np.sum(probs.T/den*ws[i + 1], axis=-1)
			
			# If paucity of samples; just use the uniform filter weights
			if np.prod(np.isfinite(w)):
				ws[i] = w
				ws[i] /= np.sum(ws[i])
			else:
				ws[i] = np.ones(N)/N
		
		return ws
	
	def calc_pred(self):
		"""
		Calculate the predicted observable and E-FRET trace, using the 
		calculate posterior distributions. These are essentially point
		estimates of the distributions, along with confidence intervals.
		"""
		
		# Extracted E-FRET Signal for BFRET method with confidence intervals
		E = self.pred['E_samples']
		self.pred['E_med'] = np.median(E, axis=0)
		self.pred['E-1SD'] = np.percentile(E, 15.87, axis=0)
		self.pred['E-2SD'] = np.percentile(E, 2.27, axis=0)
		self.pred['E-3SD'] = np.percentile(E, 0.13, axis=0)
		self.pred['E+1SD'] = np.percentile(E, 84.13, axis=0)
		self.pred['E+2SD'] = np.percentile(E, 97.73, axis=0)
		self.pred['E+3SD'] = np.percentile(E, 99.87, axis=0)
		
		# I_DD, IDA, and I_AA observables to compare
		if self.a.anl_params.do_MCMC:
			p = np.median(self.p.post_p_samples, axis=0)
		else:
			p = np.exp(self.p.MAP_log_p)
		chi = np.median(self.pred['chi_samples'], axis=0)
		x = np.vstack((np.ones(self.a.num_IDD), chi)).T
		y_pred = np.einsum('kij,kj->ki', self.ssm.H(p), x)
		AT = p[1]
		f_A_params = p[2 + self.a.num_params_f_D: 
				   2 + self.a.num_params_f_D + self.a.num_params_f_A]
		IAA_pred = AT*self.a.m.f_A.func(self.a.tDD, f_A_params)
		self.pred['IDD'] = y_pred[:, 0]
		self.pred['IDA'] = y_pred[:, 1]
		self.pred['IAA'] = IAA_pred
		
		# Calculate algebraic E-FRET as a comparison
		G = self.a.data_dict['crstlk']['G']
		a = self.a.data_dict['crstlk']['a']
		d = self.a.data_dict['crstlk']['d']
		R = (self.a.IDA - d*self.a.IDD - a*IAA_pred)/self.a.IDD
		self.pred['E_FRET_corr'] = R/(R + G)*self.a.IAA[0]/IAA_pred
		
	def plot_pred(self):
		"""
		Save the predictions to file.
		"""
		
		out_dir = self.a.res_dir + '/state_dists'
		
		fig = plt.figure(figsize=(8, 4))
		plt.plot(self.a.tDD, self.a.IDD, color='b', label=r'$I_{DD}$ Meas', 
				 lw=0.5) 
		plt.plot(self.a.tDD, self.a.IDA, color='r', label=r'$I_{DA}$ Meas', 
				 lw=0.5) 
		plt.plot(self.a.tAA, self.a.IAA, color='g', label=r'$I_{AA}$ Meas', 
				 lw=0.5)
		plt.fill_between(self.a.tDD, self.pred['IDD'] - 2*self.a.IDD_noise_sd,
						 self.pred['IDD'] + 2*self.a.IDD_noise_sd, alpha=0.3,
						 color='b', label=r'$I_{DD}$ pred')
		plt.fill_between(self.a.tDD, self.pred['IDA'] - 2*self.a.IDA_noise_sd,
						 self.pred['IDA'] + 2*self.a.IDA_noise_sd, alpha=0.3,
						 color='r', label=r'$I_{DA}$ pred')
		IAA_interp = np.interp(self.a.tAA, self.a.tDD, self.pred['IAA'])
		plt.fill_between(self.a.tAA, IAA_interp - 2*self.a.IAA_noise_sd,
						 IAA_interp + 2*self.a.IAA_noise_sd, alpha=0.3,
						 color='g', label=r'$I_{AA}$ pred') 
		plt.tick_params(which='both', labelsize=13)
		plt.xlabel(r'Time (s)', fontsize=13)
		plt.ylabel(r'Intensities', fontsize=13)
		plt.xlim(left=0, right=None)
		plt.ylim(bottom=0, top=None)
		plt.legend()
		plt.tight_layout()
		filename = '%s/Observables.png' % out_dir
		plt.savefig(filename, dpi=600)
		
		fig = plt.figure(figsize=(8, 4))
		plt.plot(self.a.tDD, self.pred['E_med'], color='b', label='Pred')
		plt.plot(self.a.tDD, self.pred['E-1SD'], color='0.75', lw=0.5)
		plt.plot(self.a.tDD, self.pred['E-2SD'], color='0.75', lw=0.5)
		plt.plot(self.a.tDD, self.pred['E+1SD'], color='0.75', lw=0.5)
		plt.plot(self.a.tDD, self.pred['E+2SD'], color='0.75', lw=0.5)
		if hasattr(self.a, 'E_true'):
			plt.plot(self.a.tDD, self.a.E_true, color='r', label='True')
		plt.tick_params(which='both', labelsize=13)
		plt.xlabel(r'Time (s)', fontsize=13)
		plt.ylabel(r'$E = \chi/D_T$', fontsize=13)
		plt.xlim(left=0, right=None)
		plt.ylim(bottom=0, top=None)
		plt.legend()
		plt.tight_layout()
		filename = '%s/E.png' % out_dir
		plt.savefig(filename, dpi=600)
		plt.close('all')
		
	def sample_from_post_dist(self):
		"""
		Generate samples from the posterior distribution of the states, 
		for each parameter of the parameter distribution, using usual
		Bayesian inference. 
		"""
		
		# Number of parameter samples (not dimension of parameter space)
		num_ps = self.ps.shape[0]
		num_steps = self.a.num_IDD
		DTs = self.ps[:, 0]
		if self.a.anl_params.state_est_all_cores:
			num_cpus = cpu_count()
		else:
			num_cpus = 1
		
		# To hold the samples from the generated posteriors
		E_samples = np.zeros((num_ps, num_steps))
		chi_samples = np.zeros((num_ps, num_steps))
		
		timenow = time.time()
		
		# Solution is analytical for Gaussian noise; (RTS smoother)
		if self.a.anl_params.process_noise == 'Gaussian':
		
			print ('Running Kalman filter and RTS smoother...')
			pool = ProcessingPool(nodes=num_cpus)
			states = pool.map(self.single_p_integrate_Gaussian, self.ps)
			post_ms = np.array([states[i][0] for i in range(num_ps)])
			post_Ps = np.array([states[i][1] for i in range(num_ps)])
			
			for j in range(num_steps):
				chi_avgs = post_ms[:, j, 1]
				chi_stds = np.sqrt(post_Ps[:, j, 1, 1])
				chi_samples[:, j] = np.random.normal(chi_avgs, chi_stds)
				E_samples[:, j] = chi_samples[:, j]/DTs
			
		elif self.a.anl_params.process_noise == 'Non-Gaussian':
		
			# Particle filter for non-Gaussian transition model
			if self.a.anl_params.state_est_method == 'SMC':
			
				print ('Forward particle filter for Non-Gaussian model...')
				pool = ProcessingPool(nodes=num_cpus)
				xs = pool.map(self.single_p_SMC_forward, self.ps)
				xs = np.array(xs)
				print ('Backward particle filter for Non-Gaussian model...')
				pool = ProcessingPool(nodes=num_cpus)
				ws = pool.map(self.single_p_SMC_backward, self.ps, xs)
				ws = np.array([ws[i] for i in range(num_ps)])
				
				# Choose one particle for each time, each p, from SMC dist.
				# Shape of chi and w: (# param samples, time, # particles)
				chis = xs[:, :, 1]
				for i in range(num_ps):
					for j in range(num_steps):
						samp = np.random.choice(chis[i, j], p=ws[i, j])
						chi_samples[i, j] = samp
						E_samples[i, j] = samp/DTs[i]
				
			# Just directly integrate the Bayesian update equations
			elif self.a.anl_params.state_est_method == 'integrate':
			
				print ('Integrating Bayesian smoothing equations directly...')
				pool = ProcessingPool(nodes=num_cpus)
				dists = pool.map(self.single_p_integrate_non_Gaussian, self.ps)
				
				# Get sample using calculated cdf of posterior distributions
				chis = dists[0]['xs']
				rand_vals = np.random.uniform(0, 1, (len(self.ps), num_steps))
				for i in range(len(self.ps)):
					cdfs = dists[i]['cdfs']
					for j in range(num_steps):
						chis_interp = np.linspace(chis[0], chis[-1], 10000)
						cdfs_interp = np.interp(chis_interp, chis, cdfs[j])
						idx = np.where(rand_vals[i, j] <= cdfs_interp)[0][0]
						chi_samples[i, j] = chis_interp[idx]
						E_samples[i, j] = chis_interp[idx]/DTs[i]
								
		print ('Time elapsed: ', time.time() - timenow)	
		
		# Calculate predictions and CIs
		self.pred = dict()
		self.pred['E_samples'] = E_samples
		self.pred['chi_samples'] = chi_samples
		self.calc_pred()
		
		# Save results and plot the parameter distributions
		out_dir = self.a.res_dir + '/state_dists'
		filename = '%s/posterior_data.pkl' % out_dir
		pickle.dump(self.pred, open(filename, 'wb'))
		self.plot_pred()


import autograd.numpy as np
from utils import norm_pdf, lognorm_pdf, standard_uniform_pdf
from autograd.scipy.special import gamma
from scipy.stats import t

class model():
	
	class f_D():

		def __init__(self):
		
			self.p_name = ['tau_D1', 'tau_D2', 'delta_D']
			self.p_init = [100, 100, 0.5]
			self.p_lo_bnd = [0, 0, 0]
			self.p_hi_bnd = [np.inf, np.inf, 1]
		
		def func(self, t, params):
	
			return params[2]*np.exp(-t/params[0]) \
				   + (1 - params[2])*np.exp(-t/params[1])
	
	class f_A():

		def __init__(self):
		
			self.p_name = ['tau_A1', 'tau_A2', 'delta_A']
			self.p_init = [100, 100, 0.5]
			self.p_lo_bnd = [0, 0, 0]
			self.p_hi_bnd = [np.inf, np.inf, 1]
			
		def func(self, t, params):
			
			return params[2]*np.exp(-t/params[0]) \
				   + (1 - params[2])*np.exp(-t/params[1])
		
	class prior_dict(object):
		"""
		Define functional form for prior distributions for D_T, A_T, but these
		still have a hyperparameter (location) to be fixed.
		"""
	
		def func(self, anl_params):
		
			pdfs_dict = dict()
			
			# Note that the ordering here is irrelevant
			# DT, AT
			pdfs_dict['DT'] = lognorm_pdf(sigma=np.log(2)).pdf
			pdfs_dict['AT'] = lognorm_pdf(sigma=np.log(1.1)).pdf
		
			# f_D params
			pdfs_dict['tau_D1'] =  lognorm_pdf(sigma=np.log(2)).pdf
			pdfs_dict['tau_D2'] = lognorm_pdf(sigma=np.log(2)).pdf
			pdfs_dict['delta_D'] = standard_uniform_pdf().pdf
			
			# f_A params
			pdfs_dict['tau_A1'] =  lognorm_pdf(sigma=np.log(1.1)).pdf
			pdfs_dict['tau_A2'] =  lognorm_pdf(sigma=np.log(1.1)).pdf
			pdfs_dict['delta_A'] =  standard_uniform_pdf().pdf
			
			# prior on process noise
			FC = anl_params.sigma_chi_prior_FC
			mode = anl_params.sigma_chi_prior_mode
			lg = lognorm_pdf(sigma=np.log(FC), fixed_mode=mode)
			pdfs_dict['sigma_chi'] = lg.pdf
			
			return pdfs_dict


"""
Calculation of Unmixing Coefficients and G Parameter

Require series of calibrationregistrations

Based on Zal and Gascoigne, 2004, doi: 10.1529/biophysj.103.022087

"""

import numpy as np
from numpy import ma
import pandas as pd

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from sklearn.linear_model import LinearRegression

from skimage import morphology
from skimage import filters
from skimage import measure
from skimage import io

from scipy import stats

from ..utils import masking
from ..utils import plot


# crosstalk estimation
class CrossReg():
    """ Class for one 3-cube FRET method crosstalk calibration registration

    """
    def __init__(self, data_path, data_name, exp_list, reg_type, trim_frame=-1):
        self.img_name = data_name
        self.img_type = reg_type
        self.img_path = data_path
        self.img_raw = io.imread(self.img_path)[:trim_frame]

        self.D_exp = exp_list[0]
        self.A_exp = exp_list[1]

        self.DD_img = self.img_raw[:,:,:,0]  # CFP-435  DD
        self.DA_img = self.img_raw[:,:,:,1]  # YFP-435  DA
        self.AD_img = self.img_raw[:,:,:,2]  # CFP-505  AD
        self.AA_img = self.img_raw[:,:,:,3]  # YFP-505  AA

        self.DD_mean_img = np.mean(self.DD_img, axis=0)
        self.DA_mean_img = np.mean(self.DA_img, axis=0)
        self.AD_mean_img = np.mean(self.AD_img, axis=0)
        self.AA_mean_img = np.mean(self.AA_img, axis=0)

        if self.img_type == 'A':
            raw_mask = self.AA_mean_img > filters.threshold_otsu(self.AA_mean_img)
        elif self.img_type == 'D':
            raw_mask = self.DD_mean_img > filters.threshold_otsu(self.DD_mean_img)

        self.mask = morphology.closing(raw_mask, footprint=morphology.disk(10))
        self.mask = morphology.erosion(self.mask, footprint=morphology.disk(10))
        self.label = measure.label(self.mask)

        self.DD_img = masking.background_extraction_along_frames(self.DD_img, self.mask)
        self.DA_img = masking.background_extraction_along_frames(self.DA_img, self.mask)
        self.AD_img = masking.background_extraction_along_frames(self.AD_img, self.mask)
        self.AA_img = masking.background_extraction_along_frames(self.AA_img, self.mask)


    def cross_calc(self):
        if self.img_type == 'A':
            _, self.a_prof_arr = masking.label_prof_arr(input_label=self.label,
                                                        input_img_series=self.DA_img / self.AA_img)
            self.a_prof_mean = np.mean(self.a_prof_arr, axis=0)
            _, self.b_prof_arr = masking.label_prof_arr(input_label=self.label,
                                                        input_img_series=self.AD_img / self.AA_img)
            self.b_prof_mean = np.mean(self.b_prof_arr, axis=0)
            # self.a_prof_mean = np.asarray([np.mean(self.DA_img[i] / self.AA_img[i]) for i in range(0, self.img_raw.shape[0])])
            # self.b_prof_mean = np.asarray([np.mean(self.AD_img[i] / self.AA_img[i]) for i in range(0, self.img_raw.shape[0])])

            self.a = np.mean(self.a_prof_mean)
            self.a_sd = np.std(self.a_prof_mean)

            self.b = np.mean(self.b_prof_mean)
            self.b_sd = np.std(self.b_prof_mean)

            a_df = pd.DataFrame({'ID':np.full(len(self.a_prof_mean), self.img_name),
                                 'type':np.full(len(self.a_prof_mean), self.img_type),
                                 'A_exp':np.full(len(self.a_prof_mean), self.A_exp),
                                 'D_exp':np.full(len(self.a_prof_mean), self.D_exp),
                                 'frame':range(len(self.a_prof_mean)),
                                 'coef':np.full(len(self.a_prof_mean), 'a'),
                                 'val':self.a_prof_mean})
            b_df = pd.DataFrame({'ID':np.full(len(self.b_prof_mean), self.img_name),
                                 'type':np.full(len(self.b_prof_mean), self.img_type),
                                 'A_exp':np.full(len(self.b_prof_mean), self.A_exp),
                                 'D_exp':np.full(len(self.b_prof_mean), self.D_exp),
                                 'frame':range(len(self.b_prof_mean)),
                                 'coef':np.full(len(self.b_prof_mean), 'b'),
                                 'val':self.b_prof_mean})
            self.cross_raw_df = pd.concat([a_df, b_df], ignore_index=True)

            self.cross_df = pd.DataFrame({'ID':[self.img_name, self.img_name],
                                          'type':[self.img_type, self.img_type],
                                          'A_exp':[self.A_exp, self.A_exp],
                                          'D_exp':[self.D_exp, self.D_exp],
                                          'coef':['a', 'b'],
                                          'val':[self.a, self.b],
                                          'sd':[self.a_sd, self.b_sd]})

        elif self.img_type == 'D':
            _, self.c_prof_arr = masking.label_prof_arr(input_label=self.label,
                                                        input_img_series=self.AA_img / self.DD_img)
            self.c_prof_mean = np.mean(self.c_prof_arr, axis=0)
            _, self.d_prof_arr = masking.label_prof_arr(input_label=self.label,
                                                        input_img_series=self.DA_img / self.DD_img)
            self.d_prof_mean = np.mean(self.d_prof_arr, axis=0)

            # self.c_prof_mean = np.asarray([np.mean(self.AA_img[i] / self.DD_img[i]) \
            #                          for i in range(0, self.img_raw.shape[0])])
            # self.d_prof_mean = np.asarray([np.mean(self.DA_img[i] / self.DD_img[i]) \
            #                          for i in range(0, self.img_raw.shape[0])])

            self.c = np.mean(self.c_prof_mean)
            self.c_sd = np.std(self.c_prof_mean)

            self.d = np.mean(self.d_prof_mean)
            self.d_sd = np.std(self.d_prof_mean)

            c_df = pd.DataFrame({'ID':np.full(len(self.c_prof_mean), self.img_name),
                                 'type':np.full(len(self.c_prof_mean), self.img_type),
                                 'A_exp':np.full(len(self.c_prof_mean), self.A_exp),
                                 'D_exp':np.full(len(self.c_prof_mean), self.D_exp),
                                 'frame':range(len(self.c_prof_mean)),
                                 'coef':np.full(len(self.c_prof_mean), 'c'),
                                 'val':self.c_prof_mean})
            d_df = pd.DataFrame({'ID':np.full(len(self.d_prof_mean), self.img_name),
                                 'type':np.full(len(self.d_prof_mean), self.img_type),
                                 'A_exp':np.full(len(self.d_prof_mean), self.A_exp),
                                 'D_exp':np.full(len(self.d_prof_mean), self.D_exp),
                                 'frame':range(len(self.d_prof_mean)),
                                 'coef':np.full(len(self.d_prof_mean), 'd'),
                                 'val':self.d_prof_mean})
            self.cross_raw_df = pd.concat([c_df, d_df], ignore_index=True)

            self.cross_df = pd.DataFrame({'ID':[self.img_name, self.img_name],
                                          'type':[self.img_type, self.img_type],
                                          'A_exp':[self.A_exp, self.A_exp],
                                          'D_exp':[self.D_exp, self.D_exp],
                                          'coef':['c', 'd'],
                                          'val':[self.c, self.d],
                                          'sd':[self.c_sd, self.d_sd]})


    def plot_hist(self):
        plt.figure(figsize=(8,8))

        ax0 = plt.subplot(211)
        ax0.hist(self.DD_mean_img.ravel(), bins=256,
                 alpha=.5,label='Ch 0 (CFP-435)', color='r')
        ax0.hist(self.DA_mean_img.ravel(), bins=256,
                 alpha=.5, label='Ch 1 (YFP-435)', color='g')
        ax0.hist(self.AD_mean_img.ravel(), bins=256,
                 alpha=.5, label='Ch 2 (CFP-505)', color='y')
        ax0.hist(self.AA_mean_img.ravel(), bins=256,
                 alpha=.5, label='Ch 3 (YFP-505)', color='b')
        ax0.legend()

        plt.title(f'File {self.img_name}, type {self.img_type}')
        plt.tight_layout()
        plt.show()


    def ff_profile(self):
        plt.figure(figsize=(8,4))

        ax0 = plt.subplot()
        ax0.plot(np.mean(self.DD_img, axis=(1,2)),
                 label='Ch 0 (CFP-435)', color='r')
        ax0.plot(np.mean(self.DA_img, axis=(1,2)),
                 label='Ch 1 (YFP-435)', color='g')
        ax0.plot(np.mean(self.AD_img, axis=(1,2)),
                 label='Ch 2 (CFP-505)', color='y')
        ax0.plot(np.mean(self.AA_img, axis=(1,2)),
                 label='Ch 3 (YFP-505)', color='b')
        ax0.legend()

        plt.title(f'File {self.img_name}, type {self.img_type}')
        plt.tight_layout()
        plt.show()        


    def ch_pic(self):
        int_min = np.min(self.img_raw)
        int_max = np.max(self.img_raw)


        plt.figure(figsize=(10,10))

        ax0 = plt.subplot(221)
        ax0.set_title('DD (Ch.0)')
        img0 = ax0.imshow(self.DD_mean_img, cmap='jet')
        img0.set_clim(vmin=int_min, vmax=int_max)
        div0 = make_axes_locatable(ax0)
        cax0 = div0.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img0, cax=cax0)
        ax0.axis('off')

        ax1 = plt.subplot(222)
        ax1.set_title('DA (Ch.1)')
        img1 = ax1.imshow(self.DA_mean_img, cmap='jet')
        img1.set_clim(vmin=int_min, vmax=int_max)
        div1 = make_axes_locatable(ax1)
        cax1 = div1.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img1, cax=cax1)
        ax1.axis('off')

        ax2 = plt.subplot(223)
        ax2.set_title('AD (Ch.2)')
        img2 = ax2.imshow(self.AD_mean_img, cmap='jet')
        img2.set_clim(vmin=int_min, vmax=int_max)
        div2 = make_axes_locatable(ax2)
        cax2 = div2.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img2, cax=cax2)
        ax2.axis('off')

        ax3 = plt.subplot(224)
        ax3.set_title('AA (Ch.3)')
        img3 = ax3.imshow(self.AA_mean_img, cmap='jet')
        img3.set_clim(vmin=int_min, vmax=int_max)
        div3 = make_axes_locatable(ax3)
        cax3 = div3.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img3, cax=cax3)
        ax3.axis('off')

        plt.suptitle(f'File {self.img_name}, type {self.img_type}')
        plt.tight_layout()
        plt.show()


class CrossRegSet():
    """ Class processing set of 3-cube FRET method crosstalk calibration registrations

    Requires sets of (A) and (D) registrations

    """
    def __init__(self, data_path, donor_reg_dict, acceptor_reg_dict, trim_frame=-1):
        self.donor_reg = donor_reg_dict
        self.acceptor_reg = acceptor_reg_dict


        self.donor_reg_list = []
        for reg_name in self.donor_reg.keys():
            name_path = data_path + f'{reg_name}.tif'
            self.donor_reg_list.append(CrossReg(data_path=name_path,
                                        data_name=reg_name,
                                        exp_list=self.donor_reg[reg_name],
                                        reg_type='D',
                                        trim_frame=trim_frame))
        self.acceptor_reg_list = []
        for reg_name in self.acceptor_reg.keys():
            name_path = data_path + f'{reg_name}.tif'
            self.acceptor_reg_list.append(CrossReg(data_path=name_path,
                                                   data_name=reg_name,
                                                   exp_list=self.acceptor_reg[reg_name],
                                                   reg_type='A',
                                                   trim_frame=trim_frame))


    def get_abcd(self, show_pic=False):
        """ Create data frame with calculated crosstalc coeficients for each calibration registration

        """

        self.cross_raw_df = pd.DataFrame(columns=['ID',  # df with raw results
                                                   'type',
                                                   'A_exp',
                                                   'D_exp',
                                                   'frame',
                                                   'coef',
                                                   'val'])

        self.cross_df = pd.DataFrame(columns=['ID',
                                                  'type',
                                                  'A_exp',
                                                  'D_exp',
                                                  'coef',
                                                  'val',
                                                  'sd'])

        for reg in self.acceptor_reg_list + self.donor_reg_list:
            if show_pic:
                reg.ch_pic()
            reg.cross_calc()
            self.cross_raw_df = pd.concat([self.cross_raw_df, reg.cross_raw_df], ignore_index=True)
            self.cross_df = pd.concat([self.cross_df, reg.cross_df], ignore_index=True)

        return self.cross_df
    

# G parameter estimation
class GReg():
    def __init__(self, img_name, raw_path, bleach_path, bleach_frame, bleach_exp, A_exp, D_exp, coef_list, trim_frame=-1):
        self.img_name = img_name
        self.img_raw = io.imread(raw_path)[:trim_frame]
        self.img_bleach = io.imread(bleach_path)[:trim_frame]

        self.bleach_frame = bleach_frame
        self.bleach_exp = bleach_exp
        self.A_exp = A_exp
        self.D_exp = D_exp

        self.a = coef_list[0]
        self.b = coef_list[1]
        self.c = coef_list[2]
        self.d = coef_list[3]


        self.DD_img = self.img_raw[:,:,:,0]  # CFP-435  DD
        self.DA_img = self.img_raw[:,:,:,1]  # YFP-435  DA
        self.AD_img = self.img_raw[:,:,:,2]  # CFP-505  AD
        self.AA_img = self.img_raw[:,:,:,3]  # YFP-505  AA

        self.DD_img_post = self.img_bleach[:,:,:,0]
        self.DA_img_post = self.img_bleach[:,:,:,1]
        self.AD_img_post = self.img_bleach[:,:,:,2]
        self.AA_img_post = self.img_bleach[:,:,:,3]


        self.DD_mean_img = np.mean(self.DD_img, axis=0)
        self.DA_mean_img = np.mean(self.DA_img, axis=0)
        self.AD_mean_img = np.mean(self.AD_img, axis=0)
        self.AA_mean_img = np.mean(self.AA_img, axis=0)

        raw_mask = self.AA_mean_img > filters.threshold_otsu(self.AA_mean_img)

        self.filtered_mask = morphology.closing(raw_mask, footprint=morphology.disk(10))
        self.back_mask = morphology.dilation(self.filtered_mask, morphology.disk(10))
        self.mask = morphology.erosion(self.filtered_mask, footprint=morphology.disk(10))
        self.label = measure.label(self.mask)

        # self.DD_img = masking.background_extraction_along_frames(self.DD_img, self.mask)
        # self.DA_img = masking.background_extraction_along_frames(self.DA_img, self.mask)
        # self.AD_img = masking.background_extraction_along_frames(self.AD_img, self.mask)
        # self.AA_img = masking.background_extraction_along_frames(self.AA_img, self.mask)

        # self.DD_img_post = masking.background_extraction_along_frames(self.DD_img_post, self.mask)
        # self.DA_img_post = masking.background_extraction_along_frames(self.DA_img_post, self.mask)
        # self.AD_img_post = masking.background_extraction_along_frames(self.AD_img_post, self.mask)
        # self.AA_img_post = masking.background_extraction_along_frames(self.AA_img_post, self.mask)

        self.Fc_pre = self.__Fc_img(dd_img=self.DD_img,
                                    da_img=self.DA_img,
                                    aa_img=self.AA_img,
                                    a=self.a, b=self.b, c=self.c, d=self.d)
        self.Fc_post = self.__Fc_img(dd_img=self.DD_img_post,
                                     da_img=self.DA_img_post, aa_img=self.AA_img_post,
                                     a=self.a, b=self.b, c=self.c, d=self.d)
        
        self.G_img = self.__G_img(Fc_pre_img=self.Fc_pre, Fc_post_img=self.Fc_post,
                                  dd_pre_img=self.DD_img, dd_post_img=self.DD_img_post,
                                  mask=self.mask)


    def G_calc_frame(self, calc_win:list[int,int]=[10,-10]):        
        _,self.G_prof_arr = masking.label_prof_arr(input_label=self.label,
                                                   input_img_series=self.G_img)

        self.G_profile = np.mean(self.G_prof_arr, axis=0)  # np.mean(self.G_img, axis=(1,2), where=self.mask)
        self.G_profile_sd = np.std(self.G_prof_arr, axis=0)  # np.std(self.G_img, axis=(1,2), where=self.mask)
        self.G = np.mean(self.G_profile[calc_win[0]:calc_win[1]])
        self.G_sd = np.std(self.G_profile[calc_win[0]:calc_win[1]])

        self.G_df = pd.DataFrame({'ID':self.img_name,
                                  'A_exp':self.A_exp,
                                  'D_exp':self.D_exp,
                                  'val':self.G,
                                  'sd':self.G_sd}, index=[0])
        

    def G_calc_fit(self, bad_cells):
        # https://realpython.com/linear-regression-in-python/)
        DD_delta_arr = np.asarray([])
        Fc_delta_arr = np.asarray([])
        for label_num in range(1, np.max(self.label)+1):
            if label_num in bad_cells:
                continue
            else:
                label_mask = self.label == label_num

                DD_delta = np.mean(self.DD_img_post, axis=(1,2), where=label_mask) - \
                        np.mean(self.DD_img, axis=(1,2), where=label_mask)
                
                Fc_delta = np.mean(self.Fc_pre, axis=(1,2), where=label_mask) - \
                        np.mean(self.Fc_post, axis=(1,2), where=label_mask)
                
                DD_delta_arr = np.concatenate((DD_delta_arr, DD_delta))
                Fc_delta_arr = np.concatenate((Fc_delta_arr, Fc_delta))


        slope, intercept, r, p, std_err = stats.linregress(DD_delta_arr, Fc_delta_arr)
        lin_mod = lambda x: slope*x + intercept

        plt.figure(figsize=(10,4))
        plt.scatter(DD_delta_arr, Fc_delta_arr, alpha=.5)
        plt.plot(DD_delta_arr, list(map(lin_mod, DD_delta_arr)),
                 color='k', linestyle='--')
        plt.xlabel('Δ DD, a.u.')
        plt.ylabel('Δ Fc, a.u.')
        plt.title(f'Registration {self.img_name}, G={round(slope,3)}+/-{round(std_err,3)} (R^2={round(r,2)}, inter.={round(intercept,2)})')
        plt.tight_layout()
        plt.show()


    def Fc_DD_pic(self):
        plt.figure(figsize=(10,4))
        for label_num in range(1, np.max(self.label)+1):
            label_mask = self.label == label_num

            # DD_delta = np.mean(self.DD_img_post - self.DD_img,
            #                    axis=(1,2), where=label_mask)
            
            DD_delta = np.mean(self.DD_img_post, axis=(1,2), where=label_mask) - \
                       np.mean(self.DD_img, axis=(1,2), where=label_mask)
            
            Fc_delta = np.mean(self.Fc_pre, axis=(1,2), where=label_mask) - \
                       np.mean(self.Fc_post, axis=(1,2), where=label_mask)

            plt.scatter(x=DD_delta,y=Fc_delta, label=label_num, alpha=.5)
        plt.xlabel('Δ DD, a.u.')
        plt.ylabel('Δ Fc, a.u.')
        plt.legend()
        plt.title(f'Registration {self.img_name}')
        plt.tight_layout()
        plt.show()


    @staticmethod
    def __Fc_img(dd_img, da_img, aa_img, a, b, c, d):
        Fc_img = []
        for frame_num in range(dd_img.shape[0]):
            DD_frame = dd_img[frame_num]
            DA_frame = da_img[frame_num]
            AA_frame = aa_img[frame_num]

            Fc_frame = DA_frame - a*(AA_frame - c*DD_frame) - d*(DD_frame - b*AA_frame)
            Fc_img.append(Fc_frame)

        return np.asarray(Fc_img)
    

    @staticmethod
    def __G_img(Fc_pre_img, Fc_post_img, dd_pre_img, dd_post_img, mask):
        G_img = []
        for frame_num in range(Fc_pre_img.shape[0]):
            Fc_pre_frame = ma.masked_where(~mask, Fc_pre_img[frame_num])
            Fc_post_frame = ma.masked_where(~mask, Fc_post_img[frame_num])
            DD_pre_frame = ma.masked_where(~mask, dd_pre_img[frame_num])
            DD_post_frame = ma.masked_where(~mask, dd_post_img[frame_num])

            G_frame = (Fc_pre_frame - Fc_post_frame) / (DD_post_frame - DD_pre_frame)
            G_img.append(G_frame)
        
        return np.asarray(G_img)


    def Fc_plot(self, raw_frames=-1, post_frames=-1):
        Fc_pre_mean = np.mean(self.Fc_pre[:raw_frames], axis=0)
        Fc_post_mean = np.mean(self.Fc_post[:post_frames], axis=0)

        plt.figure(figsize=(10,5))

        ax0 = plt.subplot(221)
        ax0.set_title('Fc pre mean')
        img0 = ax0.imshow(ma.masked_where(~self.mask, Fc_pre_mean), cmap='jet')
        # img0.set_clim(vmin=int_min, vmax=int_max)
        div0 = make_axes_locatable(ax0)
        cax0 = div0.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img0, cax=cax0)
        ax0.axis('off')

        ax1 = plt.subplot(222)
        ax1.set_title('Fc pre FF int profile')
        ax1.plot(np.mean(self.Fc_pre, axis=(1,2), where=self.mask),
                 label='Fc', color='m')
        ax1.plot(np.mean(self.DA_img, axis=(1,2), where=self.mask),
                 label='DA', color='g', linestyle='--')
        ax1.plot(np.mean(self.DD_img, axis=(1,2), where=self.mask),
                 label='DD', color='r', linestyle=':')
        ax1.legend()

        ax2 = plt.subplot(223)
        ax2.set_title('Fc post mean')
        img2 = ax2.imshow(ma.masked_where(~self.mask, Fc_post_mean), cmap='jet')
        # img0.set_clim(vmin=int_min, vmax=int_max)
        div2 = make_axes_locatable(ax2)
        cax2 = div2.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img2, cax=cax2)
        ax2.axis('off')

        ax3 = plt.subplot(224)
        ax3.set_title('Fc post FF int profile')
        ax3.plot(np.mean(self.Fc_post, axis=(1,2), where=self.mask),
                 label='Fc', color='m')
        ax3.plot(np.mean(self.DA_img_post, axis=(1,2), where=self.mask),
                 label='DA', color='g', linestyle='--')
        ax3.plot(np.mean(self.DD_img_post, axis=(1,2), where=self.mask),
                 label='DD', color='r', linestyle=':')
        ax3.legend()

        plt.suptitle(f'Registration {self.img_name}')
        plt.tight_layout()
        plt.show()
        

    def G_plot(self):
        G_mean = ma.masked_where(~self.mask, np.mean(self.G_img, axis=0))
        plt.figure(figsize=(10,10))

        ax0 = plt.subplot()
        ax0.set_title('G mean')
        img0 = ax0.imshow(G_mean, cmap='jet')
        div0 = make_axes_locatable(ax0)
        cax0 = div0.append_axes('right', size='3%', pad=0.1)
        ax0.axis('off')

        for region in measure.regionprops(self.label):
            ax0.text(region.centroid[1], region.centroid[0], region.label, color='white', fontsize=20)

        plt.colorbar(img0, cax=cax0)
        plt.tight_layout()
        plt.show()    


    def pre_post_plot(self):
        plt.figure(figsize=(10,5))

        ax0 = plt.subplot(221)
        ax0.set_title('DD (Ch.0)')
        ax0.plot(np.mean(self.DD_img, axis=(1,2), where=self.mask),
                 label='pre', color='r')
        ax0.plot(np.mean(self.DD_img_post, axis=(1,2), where=self.mask),
                 label='post', color='r', linestyle='--')
        ax0.legend()

        ax1 = plt.subplot(222)
        ax1.set_title('DA (Ch.1)')
        ax1.plot(np.mean(self.DA_img, axis=(1,2), where=self.mask),
                 label='pre', color='g')
        ax1.plot(np.mean(self.DA_img_post, axis=(1,2), where=self.mask),
                 label='post', color='g', linestyle='--')
        ax1.legend()

        ax2 = plt.subplot(223)
        ax2.set_title('AD (Ch.2)')
        ax2.plot(np.mean(self.AD_img, axis=(1,2), where=self.mask),
                 label='pre', color='y')
        ax2.plot(np.mean(self.AD_img_post, axis=(1,2), where=self.mask),
                 label='post', color='y', linestyle='--')
        ax2.legend()

        ax3 = plt.subplot(224)
        ax3.set_title('AA (Ch.3)')
        ax3.plot(np.mean(self.AA_img, axis=(1,2), where=self.mask),
                 label='pre', color='b')
        ax3.plot(np.mean(self.AA_img_post, axis=(1,2), where=self.mask),
                 label='post', color='b', linestyle='--')
        ax3.legend()
        
        plt.suptitle(f'Registration {self.img_name}')
        plt.tight_layout()
        plt.show()


    def ch_pic(self):
        int_min = np.min(self.img_raw)
        int_max = np.max(self.img_raw)


        plt.figure(figsize=(10,10))

        ax0 = plt.subplot(221)
        ax0.set_title('DD (Ch.0)')
        img0 = ax0.imshow(self.DD_mean_img, cmap='jet')
        img0.set_clim(vmin=int_min, vmax=int_max)
        div0 = make_axes_locatable(ax0)
        cax0 = div0.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img0, cax=cax0)
        ax0.axis('off')

        ax1 = plt.subplot(222)
        ax1.set_title('DA (Ch.1)')
        img1 = ax1.imshow(self.DA_mean_img, cmap='jet')
        img1.set_clim(vmin=int_min, vmax=int_max)
        div1 = make_axes_locatable(ax1)
        cax1 = div1.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img1, cax=cax1)
        ax1.axis('off')

        ax2 = plt.subplot(223)
        ax2.set_title('AD (Ch.2)')
        img2 = ax2.imshow(self.AD_mean_img, cmap='jet')
        img2.set_clim(vmin=int_min, vmax=int_max)
        div2 = make_axes_locatable(ax2)
        cax2 = div2.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img2, cax=cax2)
        ax2.axis('off')

        ax3 = plt.subplot(224)
        ax3.set_title('AA (Ch.3)')
        img3 = ax3.imshow(self.AA_mean_img, cmap='jet')
        img3.set_clim(vmin=int_min, vmax=int_max)
        div3 = make_axes_locatable(ax3)
        cax3 = div3.append_axes('right', size='3%', pad=0.1)
        plt.colorbar(img3, cax=cax3)
        ax3.axis('off')

        plt.suptitle(f'Registration {self.img_name}')
        plt.tight_layout()
        plt.show()


    def G_plot_by_label(self):
        plt.figure(figsize=(10,4))
        for label_num in range(1, np.max(self.label)+1):
            label_mask = self.label == label_num

            label_prof_mean = np.mean(self.G_img, axis=(1,2), where=label_mask)
            label_prof_sd = np.std(self.G_img, axis=(1,2), where=label_mask)

            plt.errorbar(list(range(label_prof_mean.shape[0])), label_prof_mean,
                        yerr = label_prof_sd,
                        fmt ='-o', capsize=2, label=label_num, alpha=.75)
            plt.hlines(y=np.median(label_prof_mean),
                                xmin=0, xmax=label_prof_mean.shape[0],
                                linestyles='--')

        plt.legend()
        plt.tight_layout()
        plt.show()


class GRegSet():
        def __init__(self, data_path, tandem_reg_dict, **kwargs):
            self.tandem_reg_list = []  # [raw reg, post reg, bleach frames, bleach exp, 435 exp, 505 exp]
            for reg_name in tandem_reg_dict.keys():
                reg_params = tandem_reg_dict[reg_name]
                raw_path = data_path + f'{reg_params[0]}.tif'
                bleach_path = data_path + f'{reg_params[1]}.tif'
                self.tandem_reg_list.append(GReg(img_name=reg_name,
                                                 raw_path=raw_path,
                                                 bleach_path=bleach_path,
                                                 bleach_frame=reg_params[2],
                                                 bleach_exp=reg_params[3],
                                                 A_exp=reg_params[4],
                                                 D_exp=reg_params[5],
                                                 **kwargs))


        def G_calc(self, bad_label:list[int,...]):
            self.G_df =  self.get_G(reg_list=self.tandem_reg_list)


        def G_fit_pic(self):
            for reg in self.tandem_reg_list:
                reg.Fc_DD_pic()

        def G_calc_fit(self, **kwargs):
            for reg in self.tandem_reg_list:
                reg.G_calc_fit(**kwargs)


        @ staticmethod
        def get_G(reg_list):
            """ Create data frame with calculated crosstalc coeficients for each calibration registration

            """
            G_df = pd.DataFrame(columns=['ID',
                                         'A_exp',
                                         'D_exp',
                                         'val',
                                         'sd'])

            for reg in reg_list:
                reg.G_calc_frame()
                G_df = pd.concat([G_df, reg.G_df], ignore_index=True)

            return G_df


        def draw_pic(self):
            """ Return selected plotting methods results
            Requires previous execution of 'get_G'
            
            """
            for reg in self.tandem_reg_list:
                reg.Fc_plot()
                reg.pre_post_plot()
                reg.G_plot()
                reg.G_plot_by_label()


"""
Functions for work with masks/labels and multidimensional arrays using binary masks and label

"""

import numpy as np
from numpy import ma
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.animation as anm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import colors
from matplotlib.colors import LinearSegmentedColormap

import plotly.express as px

from skimage.util import montage
from skimage.filters import rank
from skimage import morphology
from skimage import exposure
from skimage import measure
from skimage import filters
from skimage import io
from skimage import transform
from skimage import registration

from scipy import ndimage
from scipy import signal
from scipy import stats
from scipy import ndimage as ndi
from scipy import optimize


def pb_exp_correction(input_img:np.ndarray, mask:np.ndarray, method:str='exp'):
    """ Image series photobleaching correction by exponential fit.
    
    Correction proceeds by masked area of interest, not the whole frame to prevent autofluorescence influence.

    Parameters
    ----------
    input_img: ndarray [t,x,y]
        input image series
    mask: ndarray [x,y]
        mask of region of interest, must be same size with image frames
    method: str
        method for correction, exponential (`exp`) or bi-exponential (`bi_exp`)

    Returns
    -------
    corrected_img: ndarray [t,x,y]
        corrected image series
    bleach_coefs: ndarray [t]
        array of correction coeficients for each frame
    r_val: float
        R-squared value of exponential fit

    """
    exp = lambda x,a,b: a * np.exp(-b * x)
    bi_exp = lambda x,a,b,c,d: (a * np.exp(-b * x)) + (c * np.exp(-d * x))

    if method == 'exp':
        func = exp
    elif method == 'bi_exp':
        func = bi_exp
    else:
        raise ValueError('Incorrect method!')

    bleach_profile = np.mean(input_img, axis=(1,2), where=mask)
    x_profile = np.linspace(0, bleach_profile.shape[0], bleach_profile.shape[0])

    popt,_ = optimize.curve_fit(func, x_profile, bleach_profile)
    bleach_fit = np.vectorize(func)(x_profile, *popt)
    bleach_coefs =  bleach_fit / bleach_fit.max()
    bleach_coefs_arr = bleach_coefs.reshape(-1, 1, 1)
    corrected_image = input_img/bleach_coefs_arr

    _,_,r_val,_,_ = stats.linregress(bleach_profile, bleach_fit)

    return corrected_image, bleach_coefs, r_val


def proc_mask(input_img:np.ndarray,
              proc_sigma:float=1.0, win_size:int=801, k_val:float=1e-4, r_val:float=0.5,
              soma_mask:bool=False, soma_th:float=.5, soma_ext:int=100,
              ext_fin_mask:bool=False, proc_ext:int=5,
              select_largest_mask:bool=False):
    """
    __NB: used by default in most registration types!__

    Mask for single neuron images with bright soma region
    with Sauvola local threshold. Func drops soma and returns the mask for processes only.

    In the original method, a threshold T is calculated
    for every pixel in the image using the following formula
    (`m(x,y)` is mean and `s(x,y)` is SD in rectangular window):

    `
    T = m(x,y) * (1 + k * ((s(x,y) / r) - 1))
    `

    Parameters
    ----------
    input_img: ndarray
        input img for mask creation,
        recomend choose max int projection of brighter channel (GFP/YFP in our case)
    proc_sigma: float
        sigma value for gaussian blur of input image
    win_size: int
        Sauvola local threshold parameter, window size specified as a single __odd__ integer
    k_val: float
        Sauvola local threshold parameter, value of the positive parameter `k`
    r_val: float  
        Sauvola local threshold parameter, value of `r`, the dynamic range of standard deviation
    soma_mask: boolean, optional
        if True brighter region on image will identified and masked as soma
    soma_th: float, optional
        soma detection threshold in % of max img int
    soma_ext: int, optional
        soma mask extension size in px
    ext_fin_mask: boolean, optional
        if `True` - final processes mask will be extended on proc_ext value
    proc_ext: int
        neuron processes mask extention value in px
    select_largest_mask: bool, optional
        if `True` - final mask will contain the largest detected element only

    Returns
    -------
    proc_mask: ndarray, dtype boolean
        neuron processes mask
    
    """
    input_img = filters.gaussian(input_img, sigma=proc_sigma)

    # processes masking    
    proc_th = filters.threshold_sauvola(input_img, window_size=win_size, k=k_val, r=r_val)
    proc_mask = input_img > proc_th

    # mask extention
    if ext_fin_mask:
        proc_dist = ndi.distance_transform_edt(~proc_mask, return_indices=False)
        proc_mask = proc_dist <= proc_ext

    # soma masking
    if soma_mask:
        soma_region = np.copy(input_img)
        soma_region = soma_region > soma_region.max() * soma_th
        soma_region = morphology.opening(soma_region, footprint=morphology.disk(5))
        soma_dist = ndi.distance_transform_edt(~soma_region, return_indices=False)
        soma_region = soma_dist < soma_ext
        proc_mask[soma_region] = 0

    # minor mask elements filtering
    if select_largest_mask:
        def largest_only(raw_mask):
            # get larger mask element
            element_label = measure.label(raw_mask)
            element_area = {element.area : element.label for element in measure.regionprops(element_label)}
            larger_mask = element_label == element_area[max(element_area.keys())]
            return larger_mask
        proc_mask = largest_only(proc_mask)

    return proc_mask


def proc_mask_otsu(input_img:np.ndarray,
              soma_mask:bool=True, soma_th:float=.5, soma_ext:int=20,
              proc_ext:int=5, ext_fin_mask:bool=True, proc_sigma:float=.5):
    """ Mask for single neuron images with bright soma region with simple Otsu thresholding.
    Fiunc drops soma and returns the mask for processes only.

    Parameters
    ----------
    input_img: ndarray
        input img for mask creation,
        recomend choose max int projection of brighter channel (YFP)
    soma_mask: boolean, optional
        if True brighter region on image will identified and masked as soma
    soma_th: float, optional
        soma detection threshold in % of max img int
    soma_ext: int, optional
        soma mask extension size in px
    proc_ext: int
        cell processes mask extention in px
    ext_fin_mask: bool, optional
        if True - final processes mask will be extended on proc_ext val
    proc_sigma: float
        sigma value for gaussian blur of input image

    Returns
    -------
    proc_mask_fin: ndarray, dtype boolean
        extented cell processes mask
    
    """
    # soma masking
    input_img = filters.gaussian(input_img, sigma=proc_sigma)
    if soma_mask:
        soma_region = np.copy(input_img)
        soma_region = soma_region > soma_region.max() * soma_th
        soma_region = morphology.opening(soma_region, footprint=morphology.disk(5))
        # soma_region = morphology.dilation(soma_region, footprint=morphology.disk(10))
        soma_dist = ndimage.distance_transform_edt(~soma_region, return_indices=False)
        soma_mask = soma_dist < soma_ext
        input_img = ma.masked_where(soma_mask, input_img)

        th = filters.threshold_otsu(input_img.compressed())
    else:
        th = filters.threshold_otsu(input_img)

    # processes masking
    proc_mask = input_img > th
    proc_mask = morphology.closing(proc_mask, footprint=morphology.disk(5))
    if ext_fin_mask:
        proc_dist = ndimage.distance_transform_edt(~proc_mask, return_indices=False)
        proc_mask_fin = proc_dist <= proc_ext
    else:
        proc_mask_fin = proc_mask
    proc_mask_fin[soma_mask] = 0

    return proc_mask_fin


def mask_connection(input_master_mask:np.ndarray, input_minor_mask:np.ndarray):
    """ Function to filter two masks by overlay.

    Could be used in stand-alone mode as a static method of the WTvsMut class.

    Parameters
    ----------
    input_master_mask: ndarray [x,y]
        boolean mask for filtering with a greater number/area of insertions,
        typically the mask of wild-type NCS insertions
    input_minor_mask: ndarray [x,y]
        boolean mask with a lower number/area of insertions,
        typically the mask of mutant NCS insertions

    Returns
    -------
    fin_mask: ndarray [x,y]
        boolean mask that includes only the elements from 'input_wt_mask'
        that overlap with the elements from 'input_mut_mask'  
    fin_label: ndarray [x,y]
        label image for `fin_mask`

    """ 
    master_label, master_num = ndi.label(input_master_mask)

    sums = ndi.sum(input_minor_mask, master_label, np.arange(master_num+1))
    connected = sums > 0
    debris_mask = connected[master_label]

    fin_mask = np.copy(input_master_mask)
    fin_mask[~debris_mask] = 0

    fin_label, fin_num = ndi.label(fin_mask)

    return fin_mask, fin_label


def mask_along_frames(series: np.ndarray, mask: np.ndarray):
    masked_series = []

    for frame in series:
        back_mean = np.mean(frame, where=mask)
        masked_frame = np.copy(frame)
        masked_frame[~mask] = back_mean / 4
        masked_series.append(masked_frame)

    return np.asarray(masked_series)


def background_extraction_along_frames(series: np.ndarray, mask: np.ndarray):
    masked_series = []

    for frame in series:
        back_mean = np.mean(frame, where=mask)
        masked_frame = np.copy(frame)
        masked_frame = masked_frame - back_mean
        masked_series.append(masked_frame)

    return np.asarray(masked_series)


def label_prof_arr(input_label: np.ndarray, input_img_series: np.ndarray,
                   f0_win:int=3):
    """ Calc labeled ROIs profiles for time series. 

    Parameters
    ----------
    input_label: ndarray [x,y]
        label image
    input_img_series: ndarray [t,x,y]
        image time series, frames must be the same size with label array
    f0_win: int, optional
        number of points from profile start to calc F0

    Returns
    -------
    prof_df_arr: ndarray [dF_value,t]
        ΔF/F profiles for each label elements
    prof_arr: ndarray [intensity_value,t]
        intensity profiles for each label elements
 
    """
    output_dict = {}
    prof_arr = []
    prof_df_arr=[]
    for label_num in np.unique(input_label)[1:]:
        region_mask = input_label == label_num
        prof = np.mean(input_img_series, axis=(1,2), where=region_mask)
        F_0 = np.mean(prof[:f0_win])
        df_prof = (prof-F_0)/F_0
        output_dict.update({label_num:[prof, df_prof]})

        prof_arr.append(prof)
        prof_df_arr.append(df_prof)
        
    
    return np.asarray(prof_df_arr), np.asarray(prof_arr)


def trans_prof_arr(input_total_mask: np.ndarray,
                   input_label: np.ndarray,
                   input_img_series: np.ndarray):
    """ Calc ratio mask int/ROIs int for time series 

    Parameters
    ----------
    input_total_mask: ndarray, dtype boolean
        cell mask
    input_label: ndarray
        ROIs label array
    input_img_series: ndarray
        Series of images as 3D array with dim. order (time,x,y),
        each frame should be same size wiht cell mask array.

    Returns
    -------
    trans_rois_arr: ndarray
        2D array with dim. order (t,translocation profile)
    prof_df_arr: ndarray
        array with ratio "all ROIs int/cell mask int" for each frame
 
    """
    trans_rois_arr = []
    for label_num in range(1, np.max(input_label)+1):
        region_mask = input_label == label_num
        prof = np.asarray([np.sum(img, where=region_mask) / np.sum(img, where=input_total_mask) \
                           for img in input_img_series])
        trans_rois_arr.append(prof)

    total_rois = input_total_mask !=0    
    trans_total_arr = [np.sum(img, where=total_rois)/np.sum(img, where=input_total_mask) \
                       for img in input_img_series]

    return np.asarray(trans_rois_arr), np.asarray(trans_total_arr)


def series_derivate(input_img:np.ndarray, mask:np.ndarray,
                    left_win=1, space:int=0, right_win:int=1):
    """ Calculation of intensity derivative profile for image series.
    
    Parameters
    ----------
    input_img: ndarray [t,x,y]
        input image series
    mask: ndarray, dtype boolean
        cell mask
    left_win: int
        number of frames for left image
    space: int
        number of skipped frames between left and right images
    left_win: int
        number of frames for left image

    Returns
    -------
    der_arr_win: ndarray [i]
        1D array of derivate intensity calculated
        for framed images (left-skip-right), 0-1 normalized
    prof_df_arr: ndarray [i]
        1D array of derivative intensity calculated frame by frame
        and smoothed with moving average (n=3), 0-1 normalized

    """
    der_img = []
    der_arr_win = []
    der_arr_point = []
    for i in range(input_img.shape[0] - (left_win+space+right_win)):
        der_frame = np.mean(input_img[i+left_win+space:i+left_win+space+right_win+1], axis=0) - np.mean(input_img[i:i+left_win+1], axis=0) 
        der_val = np.sum(np.abs(der_frame), where=mask)
        
        der_frame_point = input_img[i+1] - input_img[i]
        der_val_point = np.sum(np.abs(der_frame_point), where=mask)

        der_img.append(der_frame)
        der_arr_win.append(der_val)
        der_arr_point.append(der_val_point)

    der_img = np.asarray(der_img)

    der_arr_win = np.asarray(der_arr_win)
    der_arr_win = (der_arr_win - np.min(der_arr_win)) / (np.max(der_arr_win)-np.min(der_arr_win))
    der_arr_win = np.pad(der_arr_win, (left_win+space+right_win), constant_values=0)[:input_img.shape[0]]

    def moving_average(a, n=3):
        ret = np.cumsum(a, dtype=float)
        ret[n:] = ret[n:] - ret[:-n]
        return ret[n - 1:] / n

    der_arr_point = np.asarray(der_arr_point)
    der_arr_point = (der_arr_point - np.min(der_arr_point)) / (np.max(der_arr_point)-np.min(der_arr_point))
    der_arr_point = moving_average(der_arr_point)
    der_arr_point = np.pad(der_arr_point, 5, constant_values=0)[:input_img.shape[0]]

    return der_arr_win, der_arr_point


def label_prof_dist(input_label: np.ndarray,
                    input_img_series: np.ndarray,
                    input_dist_img: np.ndarray):
    output_dict = {}
    for label_num in range(1, np.max(input_label)+1):
        region_mask = input_label == label_num
        prof = np.asarray([np.mean(ma.masked_where(~region_mask, img)) for img in input_img_series])
        F_0 = np.mean(prof[:5])
        df_prof = (prof-F_0)/F_0
        dist = round(np.mean(ma.masked_where(~region_mask, input_dist_img)), 1)
        output_dict.update({dist:[label_num, prof, df_prof]})
    
    return output_dict


def sorted_prof_arr_calc(input_prof_dict: np.ndarray,
                         min_dF=0.25, mid_filter=True, mid_filter_div=2):
    sorted_dist = list(input_prof_dict.keys())
    sorted_dist.sort(key=float)

    prof_arr = []
    prof_id = []
    prof_dist = []
    for dist in sorted_dist:
        roi_data = input_prof_dict[dist]
        roi_id = roi_data[0]
        # roi_prof = roi_data[1]
        roi_prof_df = roi_data[2]

        if mid_filter:
            max_idx = np.argmax(roi_prof_df)
            if max_idx > len(roi_prof_df) // mid_filter_div:
                continue
            
        if np.max(roi_prof_df) > min_dF:
            prof_id.append(roi_id)
            prof_arr.append(roi_prof_df)
            prof_dist.append(dist)

    prof_id = np.asarray(prof_id)
    prof_arr = np.asarray(prof_arr)
    prof_dist = np.asarray(prof_dist)

    return prof_arr


import autograd.numpy as np
import sys
from param_dist_funcs import param_est
from state_dist_funcs import state_est
from importlib import reload 
		

class analyze_FRET_data():
	"""
	Runs all esimation routines for a batch of data.
	"""
	
	def __init__(self, data, exp_dir):
	
		self.data_dict = data
		self.exp_dir = exp_dir
		if len(self.data_dict['IDD'].shape) == 1:
			self.num_datasets = 1
			self.num_IDD = len(self.data_dict['IDD'])
			self.num_IDA = len(self.data_dict['IDA'])
			self.num_IAA = len(self.data_dict['IAA'])
		else:
			self.num_datasets = self.data_dict['IDD'].shape[0]
			self.num_IDD = self.data_dict['IDD'].shape[1]
			self.num_IDA = self.data_dict['IDA'].shape[1]
			self.num_IAA = self.data_dict['IAA'].shape[1]
			
		# Import user-defined functions and analysis params
		# This is a bit hacky, but need to append to front, so the new
		# module location is checked first, and then reload modules.
		# This allows one to run a batch of estimations from 1 wrapper.
		sys.path = [self.exp_dir] + sys.path
		import define_model_functions
		import define_analysis_params
		reload(define_model_functions)
		reload(define_analysis_params)
		
		self.m = define_model_functions.model()
		self.anl_params = define_analysis_params.anl_params()
		self.m.f_D = self.m.f_D()
		self.m.f_A = self.m.f_A()
		self.m.prior_dict = self.m.prior_dict()
		
		if self.anl_params.process_noise == 'Non-Gaussian':
			self.m.Q_tilde = self.m.Q_tilde()
		
		# Order: DT, AT, [f_D params], [f_A params], [process noise params], G
		self.param_names = ['DT', 'AT']
		self.param_names += self.m.f_D.p_name
		self.param_names += self.m.f_A.p_name
		if self.anl_params.process_noise == 'Gaussian':
			self.param_names += ['sigma_chi']
		elif self.anl_params.process_noise == 'Non-Gaussian':
			self.param_names += self.m.Q_tilde.p_name
		
		# Convenience variables to keep track of which variables is which
		self.num_params = len(self.param_names)
		self.num_params_f_D = len(self.m.f_D.p_name)
		self.num_params_f_A = len(self.m.f_A.p_name)
		if self.anl_params.process_noise == 'Non-Gaussian':
			self.num_params_Q_tilde = len(self.m.Q_tilde.p_name)
		
		
	def run(self):
		
		for i in range(self.num_datasets): 
			self.res_dir = self.exp_dir + '/data%d' % i
			if self.num_datasets > 1:
				self.IDD = self.data_dict['IDD'][i]
				self.IAA = self.data_dict['IAA'][i]
				self.IDA = self.data_dict['IDA'][i]
				self.IDD_noise_sd = self.data_dict['IDD_noise_sd'][i]
				self.IAA_noise_sd = self.data_dict['IAA_noise_sd'][i]
				self.IDA_noise_sd = self.data_dict['IDA_noise_sd'][i]
				if 'E' in self.data_dict.keys():
					self.E_true = self.data_dict['E'][i]
			else:
				self.IDD = self.data_dict['IDD']
				self.IAA = self.data_dict['IAA']
				self.IDA = self.data_dict['IDA']
				self.IDD_noise_sd = self.data_dict['IDD_noise_sd']
				self.IAA_noise_sd = self.data_dict['IAA_noise_sd']
				self.IDA_noise_sd = self.data_dict['IDA_noise_sd']
				if 'E' in self.data_dict.keys():
					self.E_true = self.data_dict['E']
			self.tDD = self.data_dict['tDD']
			self.tAA = self.data_dict['tAA']
			
			a = param_est(self)
			a.gen_prior_dist()
			a.sample_from_post_dist()
			
			b = state_est(self, a)
			b.sample_from_post_dist()
			